from .dao_base_delete import DAOBaseDelete
from .dao_base_get import DAOBaseGet
from .dao_base_list import DAOBaseList
from .dao_base_partial_of import DAOBasePartialOf
from .dao_base_partial_update import DAOBasePartialUpdate
from .dao_base_function import DAOBaseFunction


class DAOBase(
    DAOBasePartialUpdate,
    DAOBasePartialOf,
    DAOBaseList,
    DAOBaseGet,
    DAOBaseDelete,
    DAOBaseFunction,
):
    """
    DAOBase principal composto pelos mixins especializados.
    """

    pass



from typing import Dict, List

from nsj_rest_lib.entity.filter import Filter
from nsj_rest_lib.exception import NotFoundException

from .dao_base_util import DAOBaseUtil

class DAOBaseDelete(DAOBaseUtil):

    def list_ids(self, filters: Dict[str, List[Filter]]):
        """
        Lista os IDs encontrados, de acordo com os filtros recebidos.
        """

        # Retorna None, se não receber filtros
        if filters is None or len(filters) <= 0:
            return None

        # Montando uma entity fake
        entity = self._entity_class()

        # Recuperando o campo de chave primária
        pk_field = entity.get_pk_field()

        # Organizando o where dos filtros
        filters_where, filter_values_map = self._make_filters_sql(filters)

        # Montando a query
        sql = f"""
        select {pk_field} from {entity.get_table_name()} as t0 where true {filters_where}
        """

        # Executando a query
        resp = self._db.execute_query(sql, **filter_values_map)

        # Retornando em formato de lista de IDs
        if resp is None:
            return None
        else:
            return [item[pk_field] for item in resp]

    def delete(self, filters: Dict[str, List[Filter]]):
        """
        Exclui registros de acordo com os filtros recebidos.
        """

        # Retorna None, se não receber filtros
        if filters is None or len(filters) <= 0:
            raise NotFoundException(
                f"{self._entity_class.__name__} não encontrado. Filtros: {filters}"
            )

        # Montando uma entity fake
        entity = self._entity_class()

        # Organizando o where dos filtros
        filters_where, filter_values_map = self._make_filters_sql(filters, False)

        # CUIDADO PARA NÂO EXCLUIR O QUE NÃO DEVE
        if filters_where is None or filters_where.strip() == "":
            raise NotFoundException(
                f"{self._entity_class.__name__} não encontrado. Filtros: {filters}"
            )

        # Montando a query
        sql = f"""
        delete from {entity.get_table_name()} as t0 where {filters_where}
        """

        # Executando a query
        rowcount, _ = self._db.execute(sql, **filter_values_map)

        # Verificando se houve alguma deleção
        if rowcount <= 0:
            raise NotFoundException(
                f"{self._entity_class.__name__} não encontrado. Filtros: {filters}"
            )



import enum


class ConjuntoType(enum.Enum):
    PRODUTOS = 0
    UNIDADES = 1
    COMBUSTIVEIS = 2
    SERVICOS = 3
    CLASSIFICACOESPARTICIPANTES = 4
    FICHAS = 5
    CLIENTES = 6
    FORNECEDORES = 7
    TRANSPORTADORAS = 8
    VENDEDORES = 9
    SERVICOSDECATALOGOS = 1
    MODELOSCONTRATOS = 1
    TECNICOS = 1
    RUBRICAS = 1
    REPRESENTANTESCOMERCIAIS = 1
    REPRESENTANTESTECNICOS = 1
    PROSPECTS = 1



from typing import List

from nsj_gcf_utils.json_util import convert_to_dumps
from nsj_rest_lib.entity.entity_base import EntityBase
from nsj_rest_lib.settings import USE_SQL_RETURNING_CLAUSE

from .dao_base_save_by_function import DAOBaseSaveByFunction


class DAOBaseInsert(DAOBaseSaveByFunction):

    def _sql_insert_fields(
        self, entity: EntityBase, sql_read_only_fields: List[str] = []
    ) -> str:
        """
        Retorna uma tupla com duas partes: (sql_fields, sql_ref_values), onde:
        - sql_fields: Lista de campos a inserir no insert
        - sql_ref_values: Lista das referências aos campos, a inserir no insert (parte values)
        """

        sql_fields = (
            entity._sql_fields
            if entity._sql_fields
            else [
                f"{k}"
                for k in entity.__dict__
                if not callable(getattr(entity, k, None)) and not k.startswith("_")
            ]
        )

        # Building SQL fields
        fields = [
            f"{k}"
            for k in sql_fields
            if k not in sql_read_only_fields or getattr(entity, k, None) is not None
        ]
        ref_values = [
            f":{k}"
            for k in sql_fields
            if k not in sql_read_only_fields or getattr(entity, k, None) is not None
        ]

        return (", ".join(fields), ", ".join(ref_values))

    def insert(self, entity: EntityBase, sql_read_only_fields: List[str] = []):
        """
        Insere o objeto de entidade "entity" no banco de dados
        """

        # Montando as cláusulas dos campos
        sql_fields, sql_ref_values = self._sql_insert_fields(
            entity, sql_read_only_fields
        )

        # Montando a query principal
        sql = f"""
        insert into {entity.get_table_name()} (

            {sql_fields}

        ) values (

            {sql_ref_values}

        )
        """

        # Montando as cláusulas returning
        returning_fields = entity.get_insert_returning_fields()
        if (
            getattr(entity, entity.get_pk_field()) is None
            and entity.get_pk_field() not in returning_fields
        ):
            returning_fields.append(entity.get_pk_field())

        if len(returning_fields) > 0 and USE_SQL_RETURNING_CLAUSE:
            sql_returning = ", ".join(returning_fields)

            sql += "\n"
            sql += f"returning {sql_returning}"

        # Montando um dicionário com valores das propriedades
        values_map = convert_to_dumps(entity)

        # Realizando o insert no BD
        rowcount, returning = self._db.execute(sql, **values_map)

        if rowcount <= 0:
            raise Exception(
                f"Erro inserindo {entity.__class__.__name__} no banco de dados"
            )

        # Complementando o objeto com os dados de retorno
        if len(returning_fields) > 0 and USE_SQL_RETURNING_CLAUSE:
            for field in returning_fields:
                setattr(entity, field, returning[0][field])

        return entity



import re
import uuid

from typing import Dict, List, Tuple

from nsj_gcf_utils.log_time import log_time

from nsj_rest_lib.descriptor.conjunto_type import ConjuntoType
from nsj_rest_lib.entity.entity_base import EntityBase
from nsj_rest_lib.entity.filter import Filter
from nsj_rest_lib.exception import (
    AfterRecordNotFoundException,
    NotFoundException,
)
from nsj_rest_lib.util.join_aux import JoinAux
from nsj_rest_lib.util.order_spec import (
    OrderFieldSource,
    OrderFieldSpec,
)
from nsj_rest_lib.settings import get_logger

from .dao_base_search import DAOBaseSearch


class DAOBaseList(DAOBaseSearch):

    @log_time
    def list(
        self,
        after: uuid.UUID,
        limit: int,
        fields: List[str],
        order_fields: List[OrderFieldSpec] | List[str] | None,
        filters: Dict[str, List[Filter]],
        conjunto_type: ConjuntoType = None,
        conjunto_field: str = None,
        entity_key_field: str = None,
        entity_id_value: any = None,
        search_query: str = None,
        search_fields: List[str] = None,
        joins_aux: List[JoinAux] = None,
        partial_exists_clause: Tuple[str, str, str] = None,
    ) -> List[EntityBase]:
        """
        Returns a paginated entity list.
        """

        # Creating a entity instance
        entity = self._entity_class()

        raw_order_fields = order_fields
        if raw_order_fields is None:
            raw_order_fields = entity.get_default_order_fields()

        order_specs: List[OrderFieldSpec] = []
        for field in raw_order_fields:
            if isinstance(field, OrderFieldSpec):
                order_specs.append(field)
                continue

            if not isinstance(field, str):
                raise ValueError(
                    "order_fields deve ser uma lista de strings ou OrderFieldSpec."
                )

            field_clean = re.sub(
                r"\basc\b|\bdesc\b", "", field, flags=re.IGNORECASE
            ).strip()
            is_desc = bool(re.search(r"\bdesc\b", field, flags=re.IGNORECASE))
            order_specs.append(
                OrderFieldSpec(
                    column=field_clean,
                    is_desc=is_desc,
                    source=OrderFieldSource.BASE,
                    alias=None,
                )
            )

        sql_order_items: List[Tuple[str, str, bool, str]] = []
        order_fields_alias: List[str] = []
        for spec in order_specs:
            alias_resolved = self._resolve_order_alias(spec)
            param_name = self._build_order_param(alias_resolved, spec.column)
            sql_order_items.append(
                (alias_resolved, spec.column, spec.is_desc, param_name)
            )

            clause = f"{alias_resolved}.{spec.column}"
            if spec.is_desc:
                clause = f"{clause} desc"
            order_fields_alias.append(clause)

        # Resolving data to pagination
        order_map = {param: None for _, _, _, param in sql_order_items}

        if after is not None:
            try:
                if entity_key_field is None:
                    after_obj = self.get(
                        entity.get_pk_field(),
                        after,
                        fields,
                        filters.copy(),
                        conjunto_type=conjunto_type,
                        conjunto_field=conjunto_field,
                        joins_aux=joins_aux,
                        partial_exists_clause=partial_exists_clause,
                    )
                else:
                    after_obj = self.get(
                        entity_key_field,
                        entity_id_value,
                        fields,
                        filters.copy(),
                        conjunto_type=conjunto_type,
                        conjunto_field=conjunto_field,
                        joins_aux=joins_aux,
                        partial_exists_clause=partial_exists_clause,
                    )
            except NotFoundException as e:
                raise AfterRecordNotFoundException(
                    f"Identificador recebido no parâmetro after {id}, não encontrado para a entidade {self._entity_class.__name__}."
                )

            if after_obj is not None:
                for _, column, _, param_name in sql_order_items:
                    order_map[param_name] = getattr(after_obj, column, None)

        # Making default order by clause
        order_by = f"""
            {', '.join(order_fields_alias)}
        """

        # Organizando o where da paginação
        pagination_where = ""
        if after is not None:
            # Making a list of pagination condictions
            list_page_where = []
            old_specs: List[Tuple[str, str, bool, str]] = []
            for alias, column, is_desc, param_name in sql_order_items:
                # Making equals condictions
                buffer_old_fields = "true"
                for old_alias, old_column, _, old_param in old_specs:
                    buffer_old_fields += f" and {old_alias}.{old_column} = :{old_param}"

                # Making current more than condiction
                list_page_where.append(
                    f"({buffer_old_fields} and {alias}.{column} {'<' if is_desc else '>'} :{param_name})"
                )

                # Storing current field as old
                old_specs.append((alias, column, is_desc, param_name))

            # Making SQL page condiction
            pagination_where = f"""
                and (
                    false
                    or {' or '.join(list_page_where)}
                )
            """

        # Montando o filtro de search (com ilike)
        search_map, search_where = self._make_search_sql(
            search_query, search_fields, entity
        )

        # Resolvendo o join de conjuntos (se houver)
        with_conjunto = ""
        fields_conjunto = ""
        join_conjuntos = ""
        conjunto_map = {}
        if conjunto_type is not None:
            (
                join_conjuntos,
                with_conjunto,
                fields_conjunto,
                conjunto_map,
            ) = self._make_conjunto_sql(conjunto_type, entity, filters, conjunto_field)

        # Organizando o where dos filtros
        filters_where, filter_values_map = self._make_filters_sql(filters)

        # Montando a clausula dos fields vindos dos joins
        sql_join_fields, sql_join = self._make_joins_sql(joins_aux)

        if joins_aux:
            for join in joins_aux:
                for field in join.fields:
                    setattr(self._entity_class, f"{join.alias}_{field}", None)

        partial_exists_sql = ""
        if partial_exists_clause is not None:
            (
                partial_table_name,
                partial_base_field,
                partial_relation_field,
            ) = partial_exists_clause
            partial_exists_sql = f"""
            and exists (
                select 1
                from {partial_table_name} as partial_exists
                where partial_exists.{partial_relation_field} = t0.{partial_base_field}
            )
            """

        # Montando a query em si
        sql = f"""
        {with_conjunto}
        select

            {fields_conjunto}
            {self._sql_fields(fields)}
            {sql_join_fields}

        from
            {entity.get_table_name()} as t0
            {join_conjuntos}
            {sql_join}

        where
            true
            {pagination_where}
            {filters_where}
            {search_where}
            {partial_exists_sql}

        order by
            {order_by}
        """

        # Adding limit if received
        if limit is not None:
            sql += f"        limit {limit}"

        # Making the values dict
        kwargs = {**order_map, **filter_values_map, **conjunto_map, **search_map}

        # Running the SQL query
        get_logger().debug(f"[RestLib Debug] List SQL: {sql}")
        get_logger().debug(f"[RestLib Debug] List Parameters: {kwargs}")
        resp = self._db.execute_query_to_model(sql, self._entity_class, **kwargs)

        return resp



import abc
import copy
import enum

# import uuid

from typing import Any, Dict, List, Set, Union, Optional, Tuple

from nsj_audit_lib.dto.dto_auditavel import DTOAuditavel

from nsj_rest_lib.entity.entity_base import EMPTY, EntityBase
from nsj_rest_lib.descriptor.dto_aggregator import DTOAggregator
from nsj_rest_lib.descriptor.dto_one_to_one_field import DTOOneToOneField
from nsj_rest_lib.descriptor.conjunto_type import ConjuntoType
from nsj_rest_lib.descriptor.dto_field import DTOField, DTOFieldFilter
from nsj_rest_lib.util.fields_util import (
    FieldsTree,
    clone_fields_tree,
    extract_child_tree,
    merge_fields_tree,
    normalize_fields_tree,
)
from nsj_rest_lib.util.sql_utils import montar_chave_map_sql_join


class DTOBase(abc.ABC, DTOAuditavel):
    resume_fields: Set[str] = set()
    partition_fields: Set[str] = set()
    fields_map: Dict[str, DTOField] = {}
    insert_function_field_lookup: Dict[str, Tuple[str, Any]] = {}
    update_function_field_lookup: Dict[str, Tuple[str, Any]] = {}
    get_function_field_lookup: Dict[str, Tuple[str, Any]] = {}
    list_function_field_lookup: Dict[str, Tuple[str, Any]] = {}
    delete_function_field_lookup: Dict[str, Tuple[str, Any]] = {}
    list_fields_map: dict = {}
    integrity_check_fields_map: dict = {}
    left_join_fields_map: dict = {}
    left_join_fields_map_to_query: dict = {}
    sql_join_fields_map: dict = {}
    sql_join_fields_map_to_query: dict = {}
    sql_read_only_fields: list = []
    sql_no_update_fields: Set[str] = set()
    object_fields_map: Dict[str, Any] = {}
    one_to_one_fields_map: Dict[str, DTOOneToOneField] = {}
    field_filters_map: Dict[str, DTOFieldFilter]
    aggregator_fields_map: Dict[str, DTOAggregator] = {}
    # TODO Refatorar para suportar PK composto
    pk_field: str
    fixed_filters: Dict[str, Any]
    conjunto_type: ConjuntoType
    conjunto_field: str
    escape_validator: bool
    uniques: Dict[str, Set[str]]
    candidate_keys: List[str]
    search_fields: Set[str]
    data_override_group: list[str]
    data_override_fields: list[str]
    return_hidden_fields: dict[str, any] = {}

    def __init__(
        self,
        entity: Union[EntityBase, dict] = None,
        escape_validator: bool = False,
        generate_default_pk_value: bool = True,
        validate_read_only: bool = False,
        kwargs_as_entity: bool = False,
        **kwargs,
    ) -> None:
        super().__init__()

        self.escape_validator = escape_validator
        self.generate_default_pk_value = generate_default_pk_value

        if entity is None and not kwargs_as_entity:
            self._provided_fields = set(kwargs.keys())
        else:
            self._provided_fields = set()

        # Transformando a entity em dict (se houver uma entity)
        if entity is not None:
            kwargs = (
                copy.deepcopy(entity)
                if type(entity) is dict
                else copy.deepcopy(entity.__dict__)
            )

        # NOTE: Creating this function as it will be used again when parsing
        #           DTOOneToOneField
        def set_field(dto_field: DTOField, field: str) -> None:
            if (
                validate_read_only
                and dto_field.read_only
                and kwargs.get(field, None) is not None
            ):
                raise ValueError(f"O campo {field} não pode ser preenchido.")

            # Tratando do valor default
            if (
                dto_field.default_value is not None
                and kwargs.get(field, None) is None
                and (not dto_field.pk or self.generate_default_pk_value)
            ):
                default_value = dto_field.default_value
                if callable(dto_field.default_value):
                    default_value = dto_field.default_value()
                kwargs[field] = default_value

            # Verificando se é preciso converter o nome do field para o nome
            #   correspondente no Entity. E, se será preciso aplicar alguma
            #   conversão customizada (para trazer o valor do entity para o DTO)
            entity_field = field
            if entity is not None or kwargs_as_entity:
                if dto_field.entity_field is not None:
                    entity_field = dto_field.entity_field

                # Verificando se o campo carece de conversão customizada
                if dto_field.convert_from_entity is not None:
                    fields_converted = dto_field.convert_from_entity(
                        kwargs[entity_field], kwargs
                    )
                    if field not in fields_converted:
                        setattr(self, field, None)

                    for converted_key in fields_converted:
                        setattr(self, converted_key, fields_converted[converted_key])

                    return

            # Atribuindo o valor à propriedade do DTO
            if entity_field in kwargs:
                setattr(self, field, kwargs[entity_field])
            else:
                setattr(self, field, None)
            return

        # Setando os campos registrados como fields simples
        for field in self.__class__.fields_map:
            # Recuperando a configuração do campo
            aux_dto_field = self.__class__.fields_map[field]

            set_field(aux_dto_field, field)

        # Setando os campos registrados como fields de join por meio da consulta SQL
        for field in self.__class__.sql_join_fields_map:
            # Recuperando a configuração do campo
            aux_dto_field = self.__class__.sql_join_fields_map[field]

            # Verificando a conversão do nome do field para o nome correspondente no Entity (no DTO Relacionado)
            # E, se será preciso aplicar alguma conversão customizada (para trazer o valor do entity para o DTO)
            entity_field = field
            if entity is not None or kwargs_as_entity:
                if (
                    aux_dto_field.related_dto_field
                    not in aux_dto_field.dto_type.fields_map
                ):
                    continue
                alias = self.__class__.sql_join_fields_map_to_query[
                    montar_chave_map_sql_join(aux_dto_field)
                ]
                entity_field = (
                    alias.sql_alias
                    + "_"
                    + aux_dto_field.dto_type.fields_map[
                        aux_dto_field.related_dto_field
                    ].get_entity_field_name()
                )

                # Verificando se o campo carece de conversão customizada
                if aux_dto_field.convert_from_entity is not None:
                    fields_converted = aux_dto_field.convert_from_entity(
                        kwargs[entity_field], kwargs
                    )
                    if field not in fields_converted:
                        setattr(self, field, None)

                    for converted_key in fields_converted:
                        setattr(self, converted_key, fields_converted[converted_key])

                    continue

            # Atribuindo o valor à propriedade do DTO
            if entity_field in kwargs:
                setattr(self, field, kwargs[entity_field])
            else:
                setattr(self, field, None)

        # Setando os campos registrados como fields left join
        for field in self.__class__.left_join_fields_map:
            # Recuperando a configuração do campo
            aux_dto_field = self.__class__.left_join_fields_map[field]

            # Atribuindo o valor à propriedade do DTO
            if field in kwargs:
                setattr(self, field, kwargs[field])
            else:
                setattr(self, field, None)

        # Setando os campos registrados como fields object
        for field in self.__class__.object_fields_map:
            # Recuperando a configuração do campo
            aux_dto_field = self.__class__.object_fields_map[field]

            # Atribuindo o valor à propriedade do DTO
            if field in kwargs and kwargs[field] is not None:
                if not isinstance(kwargs[field], dict):
                    raise ValueError(
                        f"O campo {field} deveria ser um dicionário com os campos da classe {aux_dto_field.dto_type}."
                    )
                else:
                    setattr(self, field, aux_dto_field.expected_type(**kwargs[field]))
            else:
                setattr(self, field, None)

        for oto_key, oto_field in self.__class__.one_to_one_fields_map.items():
            field = oto_field.entity_field
            if field not in kwargs or kwargs[field] is None:
                setattr(self, oto_key, None)
                continue
            if not isinstance(kwargs[field], (dict, oto_field.expected_type)):
                set_field(oto_field.field, oto_key)
                continue

            setattr(self, oto_key, kwargs[field])
            pass

        for k, v in self.__class__.aggregator_fields_map.items():
            if k not in kwargs or kwargs[k] is None:
                if v.not_null is True:
                    raise ValueError(f"O campo {k} deve estar preenchido.")
                setattr(self, k, None)
                continue

            if isinstance(kwargs[k], v.expected_type):
                setattr(self, k, kwargs[k])
                continue

            if isinstance(kwargs[k], dict):
                setattr(self, k, v.expected_type(**kwargs[k]))
                continue

            raise ValueError(
                f"O campo {k} deveria ser um dicionário com"
                f" os campos da classe {v.expected_type}."
            )

        # Setando os campos registrados como fields de lista
        if entity is None:
            for field in self.__class__.list_fields_map:
                dto_list_field = self.__class__.list_fields_map[field]

                if field in kwargs:
                    if kwargs[field] is None:
                        continue
                    elif not isinstance(kwargs[field], list):
                        raise ValueError(
                            f"O campo {field} deveria ser uma lista do tipo {dto_list_field.dto_type}."
                        )

                    related_itens = []
                    for item in kwargs[field]:
                        # Preenchendo os campos de particionanmento, se necessário (normalmente: tenant e grupo_empresarial)
                        for partition_field in self.__class__.partition_fields:
                            if isinstance(item, dict):
                                if (
                                    (
                                        partition_field not in item
                                        or item[partition_field] is None
                                    )
                                    and partition_field
                                    in dto_list_field.dto_type.partition_fields
                                ):
                                    partition_value = getattr(self, partition_field)
                                    item[partition_field] = partition_value
                            elif isinstance(item, dto_list_field.dto_type):
                                if (
                                    not getattr(item, partition_field)
                                    and partition_field
                                    in dto_list_field.dto_type.partition_fields
                                ):
                                    partition_value = getattr(self, partition_field)
                                    setattr(item, partition_field, partition_value)
                            else:
                                raise ValueError(
                                    f"O campo {field} deveria ser uma lista do tipo {dto_list_field.dto_type}."
                                )

                        # Criando o DTO relacionado
                        if isinstance(item, dict):
                            item_dto = dto_list_field.dto_type(**item)
                        elif isinstance(item, dto_list_field.dto_type):
                            item_dto = item
                        else:
                            raise ValueError(
                                f"O campo {field} deveria ser uma lista do tipo {dto_list_field.dto_type}."
                            )

                        # Adicionando o DTO na lista do relacionamento
                        related_itens.append(item_dto)

                    setattr(self, field, related_itens)
                else:
                    setattr(self, field, None)

        # Tratando do ID automático
        # if generate_pk_uuid:
        #     if getattr(self, self.__class__.pk_field) is None:
        #         setattr(self, self.__class__.pk_field, uuid.uuid4())

    def convert_to_entity(
        self,
        entity_class: EntityBase,
        none_as_empty: bool = False,
        is_insert: bool = False,
    ) -> EntityBase:
        """
        Cria uma instância da entidade, e a popula com os dados do DTO
        corrente.

        É importante notar que as equivalências dos nomes dos campos
        são tratadas neste método.
        """

        entity: EntityBase = entity_class()

        for field, dto_field in self.__class__.fields_map.items():
            # Verificando se é preciso realizar uma tradução de nome do campo
            entity_field = field
            if dto_field.entity_field is not None:
                entity_field = dto_field.entity_field

            # Verificando se o campo existe na entity
            if not hasattr(entity, entity_field):
                continue

            # Recuperando o valor
            value = getattr(self, field)

            # Verificando se o campo é de auto incremento gerenciado pelo banco de dados, e,
            # se o valor é None pulando o campo para não entrar na query
            if (
                is_insert
                and dto_field.auto_increment is not None
                and dto_field.auto_increment.db_managed
                and value is None
            ):
                continue

            # Verificando se é necessária alguma conversão customizada
            custom_value_converted = DTOBase.custom_convert_value_to_entity(
                value,
                dto_field,
                entity_field,
                none_as_empty,
                self.__dict__,
            )
            if len(custom_value_converted) > 0:
                for key in custom_value_converted:
                    setattr(entity, key, custom_value_converted[key])
                    entity._sql_fields.append(key)
            else:
                # Convertendo o value para o correspondente nos fields
                value = DTOBase.convert_value_to_entity(
                    value,
                    dto_field,
                    none_as_empty,
                    entity_class,
                )

                # Setando na entidade
                setattr(entity, entity_field, value)
                entity._sql_fields.append(entity_field)

        for k, _ in self.__class__.aggregator_fields_map.items():
            dto = getattr(self, k)
            if dto is None:
                # NOTE: Skipping if the field was not given
                continue

            for agg_field, dto_field in dto.fields_map.items():
                entity_field = dto_field.entity_field or agg_field

                if hasattr(entity, entity_field) is False:
                    continue

                # pylint: disable-next=protected-access
                if entity_field in entity._sql_fields:
                    # NOTE: Skipping a field if it was already created
                    #           previously. This means that the field in
                    #           the base DTO is always.
                    continue

                value = getattr(dto, agg_field)
                custom_value_converted = DTOBase.custom_convert_value_to_entity(
                    value,
                    dto_field,
                    entity_field,
                    none_as_empty,
                    self.__dict__,
                )
                if len(custom_value_converted) > 0:
                    for k1, v1 in custom_value_converted.items():
                        setattr(entity, k1, v1)
                        # pylint: disable-next=protected-access
                        entity._sql_fields.append(k1)
                else:
                    val = DTOBase.convert_value_to_entity(
                        value,
                        dto_field,
                        none_as_empty,
                        entity_class,
                    )

                    setattr(entity, entity_field, val)
                    # pylint: disable-next=protected-access
                    entity._sql_fields.append(entity_field)

        return entity

    def custom_convert_value_to_entity(
        value: Any,
        dto_field: DTOField,
        entity_field: str,
        none_as_empty: bool,
        dto_values: Dict[str, any],
    ):
        # Verificando se é necessária uma conversão customizada
        retorno = {}
        if value is not None and dto_field.convert_to_entity is not None:
            fields_converted = dto_field.convert_to_entity(value, dto_values)
            if entity_field not in fields_converted:
                retorno[entity_field] = None if not none_as_empty else EMPTY

            for converted_key in fields_converted:
                value = fields_converted[converted_key]
                if value is None and none_as_empty:
                    value = EMPTY
                retorno[converted_key] = value

        return retorno

    def convert_value_to_entity(
        value: Any,
        dto_field: DTOField,
        none_as_empty: bool,
        entity_class: EntityBase,
    ) -> Any:
        # Enumerados
        if isinstance(dto_field.expected_type, enum.EnumMeta):
            try:
                # TODO Rever se não é para verificar se o valor já é do tipo enumerado, e só retornar seu value
                # (ou índice da tupla dentro do value, se houver uma tupla)
                # Porque parece que ele tenta sempre converter para o tipo esperado, mas isso pode já ter sido
                # feito no DTO
                return DTOBase._convert_enum_to_entity(value, dto_field, none_as_empty)
            except ValueError:
                # Retornando o pórpio valor, caso se deseje converter um enumerado que não seja válido
                # Isso é, aceitando enumerados inválidos (só para os filtros)
                return value.value if isinstance(value, enum.Enum) else str(value)

        # Bool to int
        entity_fields_map = getattr(entity_class, "fields_map", {})
        entity_field_name = dto_field.entity_field or dto_field.name
        if (
            isinstance(value, bool)
            and entity_field_name in entity_fields_map
            and entity_fields_map[entity_field_name].expected_type == int
        ):
            if value is None and none_as_empty:
                return EMPTY
            return 1 if value else 0

        # Convertendo None para EMPTY (se desejado)
        if value is None:
            if none_as_empty:
                return EMPTY
            else:
                return value

        return value

    def _convert_enum_to_entity(value: Any, dto_field: DTOField, none_as_empty: bool):
        # Ignorando valores nulos
        if value is None:
            if none_as_empty:
                return EMPTY
            else:
                return None

        # Construindo uma lista com os itens do enumerado esperado
        lista_enum = list(dto_field.expected_type)

        # Se o enum estiver vazio
        if len(lista_enum) <= 0:
            return None

        # Verificando o tipo dos valores do enum
        if isinstance(lista_enum[0].value, tuple):
            # Recuperando o item correspondente do enumerado
            enumerado = value
            parou = False
            if not isinstance(value, dto_field.expected_type):
                for item in dto_field.expected_type:
                    lista_valores = list(item.value)
                    for valor in lista_valores:
                        if valor == value:
                            enumerado = item
                            parou = True
                            break

                        # Se o valor for string, testa inclusive em caixa alta e baixa
                        if isinstance(value, str):
                            if valor == value.lower() or valor == value.upper():
                                enumerado = item
                                parou = True
                                break

                    if parou:
                        break

            # Convertendo do enumerado para o valor desejado na entidade
            if "get_entity_index" in enumerado.__dict__:
                tuple_index = enumerado.get_entity_index()
            else:
                tuple_index = 1

            if isinstance(enumerado.value, list) or isinstance(enumerado.value, tuple):
                return enumerado.value[tuple_index]
            else:
                return enumerado.value
        else:
            # Tentando pelo valor do próprio enum (e testando os casos, se for str)
            if isinstance(value, str):
                try:
                    return dto_field.expected_type(value)
                except ValueError:
                    try:
                        return dto_field.expected_type(value.lower())
                    except ValueError:
                        return dto_field.expected_type(value.upper())
            else:
                return dto_field.expected_type(value)

    @classmethod
    def build_default_fields_tree(cls) -> FieldsTree:
        """
        Constrói a estrutura de fields para o DTO, com base nos campos
        configurados como resumo.

        A estrutura de fields é uma árvore, onde cada chave é o nome do campo
        e o valor é um conjunto de campos que são resumos.

        :return: Uma estrutura de fields em formato de árvore.
        :rtype: FieldsTree
        """
        tree: FieldsTree = {"root": set(cls.resume_fields)}

        for field_name, descriptor in cls.list_fields_map.items():
            if len(descriptor.resume_fields_tree.get("root", set())) <= 0:
                continue

            tree["root"].add(field_name)
            tree[field_name] = clone_fields_tree(descriptor.resume_fields_tree)

        for field_name, descriptor in cls.object_fields_map.items():
            if descriptor.resume or len(descriptor.resume_fields) > 0:
                tree["root"].add(field_name)

                if len(descriptor.resume_fields_tree.get("root", set())) > 0:
                    tree[field_name] = clone_fields_tree(descriptor.resume_fields_tree)

        for field_name, descriptor in cls.aggregator_fields_map.items():
            if field_name in tree["root"]:
                tree["root"] |= descriptor.expected_type.resume_fields

        for field_name, descriptor in cls.one_to_one_fields_map.items():
            if descriptor.field.resume:
                tree["root"].add(field_name)

        return tree

    def convert_to_dict(
        self,
        fields: Optional[FieldsTree] = None,
        expands: Optional[Dict[str, Set[str]]] = None,
        just_resume: bool = False,
    ):
        """
        Converte DTO para dict
        """

        # Resolving fields to use
        if just_resume:
            fields_tree = self.__class__.build_default_fields_tree()
        else:
            if fields is None:
                fields_tree = self.__class__.build_default_fields_tree()
            else:
                fields_tree = normalize_fields_tree(fields)
                merge_fields_tree(
                    fields_tree,
                    self.__class__.build_default_fields_tree(),
                )

        if expands is None:
            expands = {"root": set()}

        # Making result maps
        result = {}

        # Converting simple fields
        for field in self.fields_map:
            if field not in fields_tree["root"]:
                continue

            result[field] = getattr(self, field)

        # Converting sql join fields
        for field in self.sql_join_fields_map:
            if field not in fields_tree["root"]:
                continue

            result[field] = getattr(self, field)

        # Converting left join fields
        for field in self.left_join_fields_map:
            if field not in fields_tree["root"]:
                continue

            result[field] = getattr(self, field)

        for field in self.object_fields_map:
            if field not in fields_tree["root"]:
                continue

            result[field] = (
                getattr(self, field).convert_to_dict(
                    extract_child_tree(fields_tree, field)
                )
                if getattr(self, field) is not None
                else None
            )

        for field, oto_field in self.one_to_one_fields_map.items():
            if field not in fields_tree["root"] or field not in expands["root"]:
                continue

            if getattr(self, field) is None:
                result[field] = None
            else:
                if isinstance(getattr(self, field), oto_field.expected_type):
                    result[field] = getattr(self, field).convert_to_dict(
                        fields_tree[field] if field in fields_tree else None,
                        expands[field] if field in expands else None,
                    )
                    pass
                pass
            pass

        for k in self.aggregator_fields_map:
            if k not in fields_tree["root"]:
                continue

            value = getattr(self, k, None)
            if value is None:
                result[k] = None
                continue
            if hasattr(value, "convert_to_dict"):
                result[k] = value.convert_to_dict(
                    extract_child_tree(fields_tree, k),
                    extract_child_tree(expands, k),
                )
            else:
                result[k] = value

        # Converting list fields
        for field in self.list_fields_map:
            if field not in fields_tree["root"]:
                continue

            # Criando o mapa de fields para a entidade aninhada
            internal_fields = extract_child_tree(fields_tree, field)
            internal_expands = extract_child_tree(expands, field)

            # Recuperando o valor do atributo sem disparar descriptor em DTO incompleto
            value = self.__dict__.get(field, None)
            if value is None:
                result[field] = None
                continue

            # Convetendo a lista de DTOs aninhados
            result[field] = [
                (
                    item.convert_to_dict(
                        clone_fields_tree(internal_fields),
                        clone_fields_tree(internal_expands),
                        just_resume,
                    )
                    if item is not None and hasattr(item, "convert_to_dict")
                    else item
                )
                for item in value
            ]

        return result

    def get_entity_field_name(self, dto_field_name: str) -> str | None:
        """
        Retorna o nome correspondente do field no entity
        (o qual é o nome do field no DTO por padrão, ou o nome que for
        passado no parâmetro "entity_field" no construtor).

        Retorna None, se o "dto_field_name" não for achado como um dos
        fields registrados.
        """

        if dto_field_name not in self.fields_map:
            return None

        return self.fields_map[dto_field_name].get_entity_field_name()

    @classmethod
    def _filter_by_fields_tree(cls, value, fields_tree):
        if not isinstance(value, dict):
            return value

        root_fields = fields_tree.get("root", set())
        result = {}
        for field in root_fields:
            if field not in value:
                continue
            field_value = value[field]
            subtree = fields_tree.get(field)
            if subtree and isinstance(field_value, dict):
                result[field] = cls._filter_by_fields_tree(field_value, subtree)
            elif subtree and isinstance(field_value, list):
                result[field] = [
                    (
                        cls._filter_by_fields_tree(item, subtree)
                        if isinstance(item, dict)
                        else item
                    )
                    for item in field_value
                ]
            else:
                result[field] = field_value
        return result

    @classmethod
    def get_params_normalizados(
        cls, body, query_args=None, path_args=None
    ) -> list | dict:
        """
        Normaliza o corpo da requisição, de acordo com as propriedades do DTO.
        """

        fields_tree = cls.build_default_fields_tree()
        if isinstance(body, list):
            normalized_body = [
                (
                    cls._filter_by_fields_tree(item, fields_tree)
                    if isinstance(item, dict)
                    else item
                )
                for item in body
            ]
        elif isinstance(body, dict):
            normalized_body = cls._filter_by_fields_tree(body, fields_tree)
        else:
            normalized_body = {}

        return normalized_body



import os
import typing as ty

from flask import request
from typing import Callable

from nsj_audit_lib.util.audit_config import AuditConfig
from nsj_gcf_utils.json_util import json_dumps
from nsj_gcf_utils.pagination_util import PaginationException
from nsj_gcf_utils.rest_error_util import format_json_error

from nsj_rest_lib.controller.controller_util import DEFAULT_RESP_HEADERS
from nsj_rest_lib.controller.route_base import RouteBase
from nsj_rest_lib.dao.dao_base import DAOBase
from nsj_rest_lib.dto.dto_base import DTOBase
from nsj_rest_lib.entity.entity_base import EntityBase
from nsj_rest_lib.exception import (
    DataOverrideParameterException,
    MissingParameterException,
    NotFoundException,
)
from nsj_rest_lib.injector_factory_base import NsjInjectorFactoryBase
from nsj_rest_lib.settings import get_logger


class GetRoute(RouteBase):
    def __init__(
        self,
        url: str,
        http_method: str,
        dto_class: DTOBase,
        entity_class: EntityBase,
        injector_factory: NsjInjectorFactoryBase = NsjInjectorFactoryBase,
        service_name: str = None,
        handle_exception: Callable = None,
        get_function_type_class: type | None = None,
        get_function_name: str | None = None,
        get_function_response_dto_class: type | None = None,
        custom_json_response: bool = False,
        audit_config: AuditConfig | None = None,
    ):
        """
        Rota de GET por ID.

        - ``get_function_type_class``: subclasse de ``GetFunctionTypeBase``
          que representa o TYPE de entrada da função PL/pgSQL. Quando
          informado, o Service chama ``_call_function_with_type`` usando
          uma instância desse FunctionType montada a partir dos
          parâmetros da requisição. Se não for informado, o GET usa o
          modo padrão (select na tabela).
        - ``get_function_name``: nome da função PL/pgSQL para GET por
          função (ex.: ``teste.api_classificacaofinanceiraget``).
        - ``get_function_response_dto_class``: DTO usado para mapear o
          retorno da função (fallback para ``dto_class``).
        """
        super().__init__(
            url=url,
            http_method=http_method,
            dto_class=dto_class,
            entity_class=entity_class,
            dto_response_class=None,
            injector_factory=injector_factory,
            service_name=service_name,
            handle_exception=handle_exception,
            audit_config=audit_config,
        )
        self._get_function_type_class = get_function_type_class
        self._get_function_name = get_function_name
        self._get_function_response_dto_class = (
            get_function_response_dto_class or dto_class
        )
        self.custom_json_response = custom_json_response

    def _get_service(self, factory: NsjInjectorFactoryBase):
        """
        Sobrescreve o _get_service padrão para permitir configurar
        o DTO de resposta de função diretamente no construtor do Service.
        """

        if self._service_name is not None:
            return factory.get_service_by_name(self._service_name)

        from nsj_rest_lib.service.service_base import ServiceBase

        return ServiceBase(
            factory,
            DAOBase(factory.db_adapter(), self._entity_class),
            self._dto_class,
            self._entity_class,
            self._dto_response_class,
            get_function_response_dto_class=self._get_function_response_dto_class,
        )

    def handle_request(
        self,
        id: str,
        query_args: dict[str, any] = None,
        body: dict[str, any] = None,
        **kwargs: ty.Any,
    ):
        """
        Tratando requisições HTTP Get para recuperar uma instância de uma entidade.
        """

        try:
            # Recuperando os parâmetros básicos
            if os.getenv("ENV", "").lower() != "erp_sql":
                args = request.args
            else:
                args = query_args

            # Tratando dos fields
            fields = args.get("fields")
            fields = RouteBase.parse_fields(self._dto_class, fields)

            expands = RouteBase.parse_expands(self._dto_class, args.get("expand"))

            partition_fields = kwargs.copy()
            # Tratando campos de particionamento
            for field in self._dto_class.partition_fields:
                value = args.get(field)
                if value is None:
                    raise MissingParameterException(field)

                partition_fields[field] = value

            # Tratando do filtro de conjunto
            if self._dto_class.conjunto_field is not None:
                value = args.get(self._dto_class.conjunto_field)
                if value is None:
                    raise MissingParameterException(field)
                elif value is not None:
                    partition_fields[self._dto_class.conjunto_field] = value

            # Tratando dos campos de data_override
            self._validade_data_override_parameters(args)

            # Construindo os objetos
            service = self._get_service(self.get_injector_factory())
            function_object = None
            if self._get_function_type_class is not None:
                params = dict(args)
                params.update(partition_fields)
                function_object = RouteBase.build_function_type_from_args(
                    self._get_function_type_class,
                    params,
                    id_value=id,
                )
            function_params = None if function_object is not None else args

            # Chamando o service (método get)
            # TODO Rever parametro order_fields abaixo
            data = service.get(
                id,
                partition_fields,
                fields,
                expands=expands,
                function_params=function_params,
                function_object=function_object,
                function_name=self._get_function_name,
                custom_json_response=self.custom_json_response,
            )

            if self.custom_json_response and self._get_function_name is not None:
                return (json_dumps(data), 200, {**DEFAULT_RESP_HEADERS})

            # Convertendo para o formato de dicionário (permitindo omitir campos do DTO)
            dict_data = data.convert_to_dict(fields, expands)

            # Retornando a resposta da requuisição
            return (json_dumps(dict_data), 200, {**DEFAULT_RESP_HEADERS})
        except MissingParameterException as e:
            get_logger().warning(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (format_json_error(e), 400, {**DEFAULT_RESP_HEADERS})
        except DataOverrideParameterException as e:
            get_logger().warning(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (format_json_error(e), 400, {**DEFAULT_RESP_HEADERS})
        except PaginationException as e:
            get_logger().warning(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (format_json_error(e), 400, {**DEFAULT_RESP_HEADERS})
        except NotFoundException as e:
            get_logger().warning(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (format_json_error(e), 404, {**DEFAULT_RESP_HEADERS})
        except Exception as e:
            get_logger().exception(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (
                    format_json_error(f"Erro desconhecido: {e}"),
                    500,
                    {**DEFAULT_RESP_HEADERS},
                )



import copy
from typing import Any, Callable, Dict, List, Set

from flask import g

from nsj_rest_lib.dao.dao_base import DAOBase
from nsj_rest_lib.dto.after_insert_update_data import AfterInsertUpdateData
from nsj_rest_lib.dto.dto_base import DTOBase
from nsj_rest_lib.entity.entity_base import EntityBase
from nsj_rest_lib.entity.filter import Filter
from nsj_rest_lib.descriptor.filter_operator import FilterOperator
from nsj_rest_lib.exception import (
    ConflictException,
    DTOListFieldConfigException,
    NotFoundException,
)
from nsj_rest_lib.service.service_base_partial_of import (
    ServiceBasePartialOf,
    PartialExtensionWriteData,
)
from nsj_rest_lib.util.fields_util import FieldsTree


class ServiceBaseSave(ServiceBasePartialOf):
    def _save(
        self,
        insert: bool,
        dto: DTOBase,
        manage_transaction: bool,
        partial_update: bool,
        relation_field_map: Dict[str, Any] = None,
        id: Any = None,
        aditional_filters: Dict[str, Any] = None,
        custom_before_insert: Callable = None,
        custom_before_update: Callable = None,
        custom_after_insert: Callable = None,
        custom_after_update: Callable = None,
        upsert: bool = False,
        retrieve_after_insert: bool = False,
        function_name: str | None = None,
        custom_json_response: bool = False,
        retrieve_fields: FieldsTree | None = None,
    ) -> DTOBase:
        try:
            received_dto = dto
            custom_response = None

            self.fill_auto_increment_fields(insert, dto)

            if manage_transaction:
                self._dao.begin()

            old_dto = None
            should_audit_outbox = self.audit_service.should_record_audit_outbox()
            if not insert and not upsert:
                old_dto = self._retrieve_old_dto(dto, id, aditional_filters)
                setattr(dto, dto.pk_field, getattr(old_dto, dto.pk_field))

            if not insert and upsert:
                old_dto = dto

            if custom_before_insert:
                received_dto = copy.deepcopy(dto)
                dto = custom_before_insert(self._dao._db, dto)

            if custom_before_update:
                if received_dto == dto:
                    received_dto = copy.deepcopy(dto)
                dto = custom_before_update(self._dao._db, old_dto, dto)

            partial_write_data: PartialExtensionWriteData | None = None
            if self._has_partial_support():
                entity, partial_write_data = self._prepare_partial_save_entities(
                    dto,
                    partial_update,
                    insert,
                )
            else:
                entity = dto.convert_to_entity(
                    self._entity_class,
                    partial_update,
                    insert,
                )

            if id is None:
                id = getattr(entity, entity.get_pk_field())

            entity_pk_field = self._entity_class().get_pk_field()
            if getattr(entity, entity_pk_field) is None and insert:
                setattr(entity, entity_pk_field, id)

            if relation_field_map is not None:
                for entity_field, value in relation_field_map.items():
                    if hasattr(entity, entity_field):
                        setattr(entity, entity_field, value)
                        if entity_field not in entity._sql_fields:
                            entity._sql_fields.append(entity_field)

            if (insert and hasattr(entity, self._created_by_property)) or (
                hasattr(entity, self._updated_by_property)
            ):
                if g and hasattr(g, "profile") and g.profile is not None:
                    auth_type_is_api_key = g.profile["authentication_type"] == "api_key"
                    user = g.profile["email"]
                    if insert and hasattr(entity, self._created_by_property):
                        if not auth_type_is_api_key:
                            setattr(entity, self._created_by_property, user)
                        else:
                            value = getattr(entity, self._created_by_property)
                            if value is None or value == "":
                                raise ValueError(
                                    f"É necessário preencher o campo '{self._created_by_property}'."
                                )
                    if hasattr(entity, self._updated_by_property):
                        if not auth_type_is_api_key:
                            setattr(entity, self._updated_by_property, user)
                        else:
                            value = getattr(entity, self._updated_by_property)
                            if value is None or value == "":
                                raise ValueError(
                                    f"É necessário preencher o campo '{self._updated_by_property}'"
                                )

            if aditional_filters is not None:
                aditional_entity_filters = self._create_entity_filters(
                    aditional_filters
                )
            else:
                aditional_entity_filters = {}

            for unique in self._dto_class.uniques:
                unique = self._dto_class.uniques[unique]
                self._check_unique(
                    dto,
                    entity,
                    aditional_entity_filters,
                    unique,
                    old_dto,
                )

            if insert:
                if self.entity_exists(entity, aditional_entity_filters):
                    raise ConflictException(
                        f"Já existe um registro no banco com o identificador '{getattr(entity, entity_pk_field)}'"
                    )

                self.audit_service.record_audit_outbox(
                    action="insert",
                    dto=dto,
                    resource_id=id,
                    old_dto=None,
                    route_resource_id=None,
                )

                ################################################
                # DAO.INSERT (ou DAO.INSERT_BY_FUNCTION)
                ################################################
                if self._insert_function_type_class is None:
                    entity = self._dao.insert(entity, dto.sql_read_only_fields)
                else:
                    insert_function_object = self._build_insert_function_type_object(
                        dto
                    )
                    custom_response = self._dao.insert_by_function(
                        insert_function_object,
                        function_name=function_name,
                        custom_json_response=custom_json_response,
                    )

                if partial_write_data is not None:
                    self._handle_partial_extension_insert(entity, partial_write_data)

                if self._dto_class.conjunto_type is not None:
                    conjunto_field_value = getattr(dto, self._dto_class.conjunto_field)

                    aditional_filters[self._dto_class.conjunto_field] = (
                        conjunto_field_value
                    )

                    self._dao.insert_relacionamento_conjunto(
                        id, conjunto_field_value, self._dto_class.conjunto_type
                    )
            else:
                if self._update_function_type_class is not None and upsert:
                    raise ValueError(
                        "update_by_function não suporta operações com upsert."
                    )

                audit_old_dto = old_dto
                if should_audit_outbox and old_dto is None:
                    audit_id = id or getattr(dto, dto.pk_field, None)
                    if audit_id is not None:
                        try:
                            audit_old_dto = self._retrieve_old_dto(
                                dto, audit_id, aditional_filters
                            )
                        except NotFoundException:
                            audit_old_dto = None

                resource_id = id or getattr(old_dto, dto.pk_field, None)
                self.audit_service.record_audit_outbox(
                    action="update",
                    dto=dto,
                    resource_id=resource_id,
                    old_dto=audit_old_dto,
                    route_resource_id=id,
                )

                if self._update_function_type_class is None:
                    entity = self._dao.update(
                        entity.get_pk_field(),
                        getattr(old_dto, dto.pk_field),
                        entity,
                        aditional_entity_filters,
                        partial_update,
                        dto.sql_read_only_fields,
                        dto.sql_no_update_fields,
                        upsert,
                    )
                else:
                    update_function_object = self._build_update_function_type_object(
                        dto
                    )
                    custom_response = self._dao.update_by_function(
                        update_function_object,
                        function_name=function_name,
                        custom_json_response=custom_json_response,
                    )

                if partial_write_data is not None:
                    self._handle_partial_extension_update(
                        entity,
                        partial_write_data,
                        partial_update,
                    )

            if (
                self._dto_post_response_class is not None
                and not retrieve_after_insert
                and not custom_json_response
            ):
                response_dto = self._dto_post_response_class(
                    entity, escape_validator=True
                )
            else:
                response_dto = None

            if len(self._dto_class.list_fields_map) > 0:
                self._save_related_lists(
                    insert, dto, entity, partial_update, response_dto, aditional_filters
                )

            if custom_after_insert is not None or custom_after_update is not None:
                new_dto = self._dto_class(entity, escape_validator=True)

                for list_field in dto.list_fields_map:
                    setattr(new_dto, list_field, getattr(dto, list_field))

                if (
                    self._dto_class.conjunto_field is not None
                    and getattr(new_dto, self._dto_class.conjunto_field) is None
                ):
                    value_conjunto = getattr(dto, self._dto_class.conjunto_field)
                    setattr(new_dto, self._dto_class.conjunto_field, value_conjunto)

            after_data = AfterInsertUpdateData()
            after_data.received_dto = received_dto

            custom_data = None
            if insert:
                if custom_after_insert is not None:
                    custom_data = custom_after_insert(
                        self._dao._db, new_dto, after_data
                    )
            else:
                if custom_after_update is not None:
                    custom_data = custom_after_update(
                        self._dao._db, old_dto, new_dto, after_data
                    )

            if retrieve_after_insert and not custom_json_response:
                response_dto = self.get(id, aditional_filters, retrieve_fields)

            if custom_data is not None:
                if isinstance(custom_data, dict):
                    if response_dto is not None:
                        for key in custom_data:
                            setattr(response_dto, key, custom_data[key])
                    else:
                        response_dto = custom_data
                else:
                    if response_dto is not None:
                        pass
                    else:
                        response_dto = custom_data

            if custom_json_response and custom_response is not None:
                return custom_response

            return response_dto
        except:
            if manage_transaction:
                self._dao.rollback()
            raise
        finally:
            if manage_transaction:
                self._dao.commit()

    def fill_auto_increment_fields(self, insert, dto):
        if insert:
            auto_increment_fields = getattr(self._dto_class, "auto_increment_fields")

            for field_key in auto_increment_fields:
                field = self._dto_class.fields_map[field_key]

                if dto.__dict__.get(field.name, None):
                    continue

                if field.auto_increment.db_managed:
                    continue

                group_fields = set(field.auto_increment.group)
                for partition_field in dto.partition_fields:
                    if partition_field not in group_fields:
                        group_fields.add(partition_field)
                group_fields = list(group_fields)
                group_fields.sort()

                group_values = []
                for group_field in group_fields:
                    group_values.append(str(getattr(dto, group_field, "----")))

                next_value = self._dao.next_val(
                    sequence_base_name=field.auto_increment.sequence_name,
                    group_fields=group_values,
                    start_value=field.auto_increment.start_value,
                )

                obj_values = {}
                for f in dto.fields_map:
                    obj_values[f] = getattr(dto, f)

                value = field.auto_increment.template.format(
                    **obj_values, seq=next_value
                )

                if field.expected_type is int:
                    setattr(dto, field.name, int(value))
                else:
                    setattr(dto, field.name, value)

    def _retrieve_old_dto(self, dto, id, aditional_filters):
        fields = self._make_fields_from_dto(dto)
        get_filters = (
            copy.deepcopy(aditional_filters) if aditional_filters is not None else {}
        )

        if (
            self._dto_class.conjunto_field is not None
            and self._dto_class.conjunto_field not in get_filters
        ):
            get_filters[self._dto_class.conjunto_field] = getattr(
                dto, self._dto_class.conjunto_field
            )

        for pt_field in dto.partition_fields:
            pt_value = getattr(dto, pt_field, None)
            if pt_value is not None:
                get_filters[pt_field] = pt_value

        old_dto = self.get(id, get_filters, fields)

        if (
            self._dto_class.conjunto_field is not None
            and getattr(old_dto, self._dto_class.conjunto_field) is None
        ):
            value_conjunto = getattr(dto, self._dto_class.conjunto_field)
            setattr(old_dto, self._dto_class.conjunto_field, value_conjunto)
        return old_dto

    def _save_related_lists(
        self,
        insert: bool,
        dto: DTOBase,
        entity: EntityBase,
        partial_update: bool,
        response_dto: DTOBase,
        aditional_filters: Dict[str, Any] = None,
    ):
        # TODO Controlar profundidade?!
        from .service_base import ServiceBase

        for master_dto_field, list_field in self._dto_class.list_fields_map.items():
            response_list = []

            detail_list = getattr(dto, master_dto_field)

            if detail_list is None:
                continue

            if (
                insert
                and self._insert_function_type_class is not None
                and getattr(list_field, "insert_function_type", None) is not None
            ):
                continue

            detail_dao = DAOBase(
                self._injector_factory.db_adapter(), list_field.entity_type
            )

            if list_field.service_name is not None:
                detail_service = self._injector_factory.get_service_by_name(
                    list_field.service_name
                )
            else:
                detail_service = ServiceBase(
                    self._injector_factory,
                    detail_dao,
                    list_field.dto_type,
                    list_field.entity_type,
                    list_field.dto_post_response_type,
                )

            relation_key_field = entity.get_pk_field()
            if list_field.relation_key_field is not None:
                relation_key_field = dto.get_entity_field_name(
                    list_field.relation_key_field
                )

            relation_key_value = getattr(entity, relation_key_field)

            relation_field_map = {
                list_field.related_entity_field: relation_key_value,
            }

            old_detail_ids = None
            if not insert:
                relation_condiction = Filter(FilterOperator.EQUALS, relation_key_value)

                relation_filter = {
                    list_field.related_entity_field: [relation_condiction]
                }

                for field in self._dto_class.partition_fields:
                    if field in list_field.dto_type.partition_fields:
                        relation_filter[field] = [
                            Filter(FilterOperator.EQUALS, getattr(dto, field))
                        ]

                old_detail_ids = detail_dao.list_ids(relation_filter)

            detail_upsert_list = []

            for detail_dto in detail_list:
                detail_pk_field = detail_dto.__class__.pk_field
                detail_pk = getattr(detail_dto, detail_pk_field)

                is_detail_insert = True
                if old_detail_ids is not None and detail_pk in old_detail_ids:
                    is_detail_insert = False
                    old_detail_ids.remove(detail_pk)

                if self._dto_class.pk_field is None:
                    raise DTOListFieldConfigException(
                        f"PK field not found in class: {self._dto_class}"
                    )

                if self._dto_class.pk_field not in dto.__dict__:
                    raise DTOListFieldConfigException(
                        f"PK field not found in DTO: {self._dto_class}"
                    )

                detail_upsert_list.append(
                    {
                        "is_detail_insert": is_detail_insert,
                        "detail_dto": detail_dto,
                        "detail_pk": detail_pk,
                    }
                )

            if (
                not partial_update
                and old_detail_ids is not None
                and len(old_detail_ids) > 0
            ):
                for old_id in old_detail_ids:
                    detail_service.delete(old_id, aditional_filters)

            for item in detail_upsert_list:
                response_detail_dto = detail_service._save(
                    item["is_detail_insert"],
                    item["detail_dto"],
                    False,
                    partial_update if not item["is_detail_insert"] else False,
                    relation_field_map,
                    item["detail_pk"],
                    aditional_filters=aditional_filters,
                )

                response_list.append(response_detail_dto)

            if (
                response_dto is not None
                and master_dto_field in response_dto.list_fields_map
                and list_field.dto_post_response_type is not None
            ):
                setattr(response_dto, master_dto_field, response_list)

    def _check_unique(
        self,
        dto: DTOBase,
        entity: EntityBase,
        entity_filters: Dict[str, List[Filter]],
        unique: Set[str],
        old_dto: DTOBase,
    ):
        unique_filter = {}
        for field in unique:
            value = getattr(dto, field)
            if value is None:
                return
            unique_filter[field] = value

        unique_entity_filters = self._create_entity_filters(unique_filter)

        if entity.get_pk_field() in unique_entity_filters:
            del unique_entity_filters[entity.get_pk_field()]

        if len(unique_entity_filters) <= 0:
            return

        entity_filters = {**entity_filters, **unique_entity_filters}

        filters_pk = entity_filters.setdefault(entity.get_pk_field(), [])
        filters_pk.append(
            Filter(
                FilterOperator.DIFFERENT,
                (
                    getattr(old_dto, dto.pk_field)
                    if old_dto is not None
                    else getattr(dto, dto.pk_field)
                ),
            )
        )

        try:
            encontrados = self._dao.list(
                None,
                1,
                [entity.get_pk_field()],
                None,
                entity_filters,
            )

            if len(encontrados) >= 1:
                raise ConflictException(
                    f"Restrição de unicidade violada para a unique: {unique}"
                )
        except NotFoundException:
            return



import os
import typing as ty

from flask import request
from typing import Callable

from nsj_audit_lib.util.audit_config import AuditConfig
from nsj_gcf_utils.json_util import json_dumps, JsonLoadException
from nsj_gcf_utils.rest_error_util import format_json_error

from nsj_rest_lib.controller.controller_util import DEFAULT_RESP_HEADERS
from nsj_rest_lib.controller.route_base import RouteBase
from nsj_rest_lib.dto.dto_base import DTOBase
from nsj_rest_lib.dto.queued_data_dto import QueuedDataDTO
from nsj_rest_lib.entity.entity_base import EntityBase
from nsj_rest_lib.exception import MissingParameterException, NotFoundException
from nsj_rest_lib.injector_factory_base import NsjInjectorFactoryBase
from nsj_rest_lib.settings import get_logger


class PatchRoute(RouteBase):
    def __init__(
        self,
        url: str,
        http_method: str,
        dto_class: DTOBase,
        entity_class: EntityBase,
        dto_response_class: DTOBase = None,
        injector_factory: NsjInjectorFactoryBase = NsjInjectorFactoryBase,
        service_name: str = None,
        handle_exception: Callable = None,
        custom_before_update: Callable = None,
        custom_after_update: Callable = None,
        retrieve_after_partial_update: bool = False,
        custom_json_response: bool = False,
        audit_config: AuditConfig | None = None,
    ):
        super().__init__(
            url=url,
            http_method=http_method,
            dto_class=dto_class,
            entity_class=entity_class,
            dto_response_class=dto_response_class,
            injector_factory=injector_factory,
            service_name=service_name,
            handle_exception=handle_exception,
            audit_config=audit_config,
        )
        self.custom_before_update = custom_before_update
        self.custom_after_update = custom_after_update
        self.retrieve_after_partial_update = retrieve_after_partial_update
        self.custom_json_response = custom_json_response

    def handle_request(
        self,
        id: str,
        query_args: dict[str, any] = None,
        body: dict[str, any] = None,
        **kwargs: ty.Any,
    ):
        """
        Tratando requisições HTTP Put para inserir uma instância de uma entidade.
        """

        try:
            # Recuperando os dados do corpo da requisição
            if os.getenv("ENV", "").lower() != "erp_sql":
                data = request.json
                args = request.args
            else:
                data = body
                args = query_args or {}

            if len(kwargs) > 0:
                data.update(kwargs)

            # Convertendo os dados para o DTO
            data = self._dto_class(
                validate_read_only=True,
                escape_validator=True,
                **data,
            )

            # Reaplicando validação apenas nos campos enviados
            data.escape_validator = False
            for field_name in getattr(data, "_provided_fields", set()):
                if field_name in data.fields_map:
                    setattr(data, field_name, getattr(data, field_name))

            # Montando os filtros de particao de dados
            partition_filters = kwargs.copy()

            for field in data.partition_fields:
                value = getattr(data, field)
                if value is None:
                    raise MissingParameterException(field)
                elif value is not None:
                    partition_filters[field] = value

            # Construindo os objetos
            service = self._get_service(self.get_injector_factory())
            retrieve_fields = (
                RouteBase.parse_fields(self._dto_class, args.get("fields"))
                if self.retrieve_after_partial_update
                else None
            )

            # Chamando o service (método insert)
            data = service.partial_update(
                dto=data,
                id=id,
                aditional_filters=partition_filters,
                custom_before_update=self.custom_before_update,
                custom_after_update=self.custom_after_update,
                retrieve_after_partial_update=self.retrieve_after_partial_update,
                custom_json_response=self.custom_json_response,
                retrieve_fields=retrieve_fields,
            )

            if data is not None:
                # Verificando se houve um enfileiramento (pelo custom_after_update)
                if isinstance(data, QueuedDataDTO):
                    queued_data: QueuedDataDTO = data
                    resp_headers = {
                        **DEFAULT_RESP_HEADERS,
                        "Location": queued_data.status_url,
                    }
                    return ("", 202, resp_headers)

                if self.custom_json_response and (
                    isinstance(data, dict)
                    or (
                        isinstance(data, list)
                        and (not data or not hasattr(data[0], "convert_to_dict"))
                    )
                ):
                    return (json_dumps(data), 200, {**DEFAULT_RESP_HEADERS})

                # Convertendo para o formato de dicionário
                dict_data = data.convert_to_dict(retrieve_fields)

                # Retornando a resposta da requuisição
                return (json_dumps(dict_data), 200, {**DEFAULT_RESP_HEADERS})
            else:
                # Retornando a resposta da requuisição
                return ("", 204, {**DEFAULT_RESP_HEADERS})
        except JsonLoadException as e:
            get_logger().warning(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (format_json_error(e), 400, {**DEFAULT_RESP_HEADERS})
        except MissingParameterException as e:
            get_logger().warning(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (format_json_error(e), 400, {**DEFAULT_RESP_HEADERS})
        except ValueError as e:
            get_logger().warning(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (format_json_error(e), 400, {**DEFAULT_RESP_HEADERS})
        except NotFoundException as e:
            get_logger().warning(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (format_json_error(e), 404, {**DEFAULT_RESP_HEADERS})
        except Exception as e:
            get_logger().exception(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (
                    format_json_error(f"Erro desconhecido: {e}"),
                    500,
                    {**DEFAULT_RESP_HEADERS},
                )



import typing as ty

from nsj_rest_lib.descriptor.dto_list_field import DTOListField
from nsj_rest_lib.descriptor.dto_object_field import DTOObjectField
from nsj_rest_lib.descriptor.dto_one_to_one_field import DTOOneToOneField
from nsj_rest_lib.dto.dto_base import DTOBase
from nsj_rest_lib.entity.function_type_base import (
    FunctionTypeBase,
    InsertFunctionTypeBase,
    UpdateFunctionTypeBase,
)


class ServiceBaseSaveByFunction:
    _insert_function_type_class: ty.Optional[ty.Type[InsertFunctionTypeBase]] = None
    _update_function_type_class: ty.Optional[ty.Type[UpdateFunctionTypeBase]] = None

    def set_insert_function_type_class(
        self,
        insert_function_type_class: ty.Optional[
            ty.Type[InsertFunctionTypeBase]
        ],
    ):
        if insert_function_type_class is not None and not issubclass(
            insert_function_type_class, InsertFunctionTypeBase
        ):
            raise ValueError(
                "A classe informada em insert_function_type_class deve herdar de InsertFunctionTypeBase."
            )

        self._insert_function_type_class = insert_function_type_class

        if (
            self._insert_function_type_class is not None
            and getattr(self, "_dto_class", None) is not None
        ):
            self._insert_function_type_class.get_function_mapping(self._dto_class)

    def set_update_function_type_class(
        self,
        update_function_type_class: ty.Optional[
            ty.Type[UpdateFunctionTypeBase]
        ],
    ):
        if update_function_type_class is not None and not issubclass(
            update_function_type_class, UpdateFunctionTypeBase
        ):
            raise ValueError(
                "A classe informada em update_function_type_class deve herdar de UpdateFunctionTypeBase."
            )

        self._update_function_type_class = update_function_type_class

        if (
            self._update_function_type_class is not None
            and getattr(self, "_dto_class", None) is not None
        ):
            self._update_function_type_class.get_function_mapping(self._dto_class)

    def _build_insert_function_type_object(
        self,
        dto: DTOBase,
        insert_function_type_class: ty.Optional[
            ty.Type[InsertFunctionTypeBase]
        ] = None,
    ):
        return self._build_function_type_object(
            dto,
            insert_function_type_class,
            self._insert_function_type_class,
            operation="insert",
        )

    def _build_update_function_type_object(
        self,
        dto: DTOBase,
        update_function_type_class: ty.Optional[
            ty.Type[UpdateFunctionTypeBase]
        ] = None,
    ):
        return self._build_function_type_object(
            dto,
            update_function_type_class,
            self._update_function_type_class,
            operation="update",
        )

    def _build_function_type_object(
        self,
        dto: DTOBase,
        override_class: ty.Optional[ty.Type[FunctionTypeBase]],
        default_class: ty.Optional[ty.Type[FunctionTypeBase]],
        operation: str,
    ):
        target_class = override_class or default_class

        if target_class is None:
            return None

        mapping_dto_class = (
            dto.__class__ if override_class is not None else self._dto_class
        )

        mapping = target_class.get_function_mapping(mapping_dto_class)

        return self._build_function_type_object_from_mapping(
            dto,
            target_class,
            mapping,
            operation=operation,
        )

    def _build_function_type_object_from_mapping(
        self,
        dto: DTOBase,
        function_type_class: ty.Type[FunctionTypeBase],
        mapping: ty.Dict[str, ty.Tuple[str, ty.Any]],
        operation: str,
    ) -> FunctionTypeBase:
        if mapping is None:
            raise ValueError(
                f"FunctionType '{function_type_class.__name__}' não possui mapeamentos configurados."
            )

        insert_object = function_type_class()
        dto_values = dto.__dict__

        for function_field_name, (dto_field_name, descriptor) in mapping.items():
            if not hasattr(dto, dto_field_name):
                raise ValueError(
                    f"DTO '{dto.__class__.__name__}' não possui o campo '{dto_field_name}' utilizado em '{function_type_class.__name__}'."
                )

            value = getattr(dto, dto_field_name, None)

            convert_to_function = getattr(descriptor, "convert_to_function", None)
            if convert_to_function is not None:
                converted_values = convert_to_function(value, dto_values) or {}

                if not isinstance(converted_values, dict):
                    raise ValueError(
                        f"A função 'convert_to_function' configurada no campo '{dto_field_name}' deve retornar um dicionário."
                    )

                if function_field_name not in converted_values:
                    converted_values = {
                        function_field_name: None,
                        **converted_values,
                    }

                for target_field, target_value in converted_values.items():
                    setattr(insert_object, target_field, target_value)
                continue

            if isinstance(
                descriptor,
                (DTOListField, DTOObjectField, DTOOneToOneField),
            ):
                relation_value = self._build_function_relation_value(
                    descriptor,
                    value,
                    operation,
                )
                if relation_value is not None:
                    setattr(insert_object, function_field_name, relation_value)
                continue

            setattr(insert_object, function_field_name, value)

        return insert_object

    def _build_function_relation_value(
        self,
        descriptor: ty.Union[DTOListField, DTOObjectField, DTOOneToOneField],
        value: ty.Any,
        operation: str,
    ):
        if value is None:
            return None

        function_type_class = descriptor.get_function_type(operation)
        if function_type_class is None:
            raise ValueError(
                f"O campo '{descriptor.name}' precisa informar 'function_type' para relacionamentos ({operation})."
            )

        dto_class = self._get_relation_dto_class(descriptor)
        mapping = function_type_class.get_function_mapping(dto_class)

        if isinstance(descriptor, DTOListField):
            related_values = []
            for item in value:
                dto_instance = self._ensure_dto_instance(item, dto_class)
                if dto_instance is None:
                    continue
                related_values.append(
                    self._build_function_type_object_from_mapping(
                        dto_instance,
                        function_type_class,
                        mapping,
                        operation=operation,
                    )
                )
            return related_values

        dto_instance = self._ensure_dto_instance(value, dto_class)
        if dto_instance is None:
            return None

        return self._build_function_type_object_from_mapping(
            dto_instance,
            function_type_class,
            mapping,
            operation=operation,
        )

    def _get_relation_dto_class(
        self, descriptor: ty.Union[DTOListField, DTOObjectField, DTOOneToOneField]
    ) -> ty.Type[DTOBase]:
        if isinstance(descriptor, DTOListField):
            return descriptor.dto_type
        return descriptor.expected_type

    def _ensure_dto_instance(
        self,
        value: ty.Any,
        dto_class: ty.Type[DTOBase],
    ) -> ty.Optional[DTOBase]:
        if value is None:
            return None

        if isinstance(value, dto_class):
            return value

        if isinstance(value, dict):
            return dto_class(**value)

        raise ValueError(
            f"O valor informado para o relacionamento deveria ser do tipo '{dto_class.__name__}'. Valor recebido: {type(value)}."
        )



import uuid

from typing import List, Tuple

from nsj_rest_lib.descriptor.conjunto_type import ConjuntoType
from nsj_rest_lib.entity.entity_base import EntityBase
from nsj_rest_lib.exception import ConflictException, NotFoundException
from nsj_rest_lib.util.join_aux import JoinAux

from .dao_base_conjuntos import DAOBaseConjuntos


class DAOBaseGet(DAOBaseConjuntos):

    def get(
        self,
        key_field: str,
        id: uuid.UUID,
        fields: List[str] = None,
        filters=None,
        conjunto_type: ConjuntoType = None,
        conjunto_field: str = None,
        joins_aux: List[JoinAux] = None,
        override_data: bool = False,
        partial_exists_clause: Tuple[str, str, str] = None,
    ) -> EntityBase:
        """
        Returns an entity instance by its ID.
        """

        # Creating a entity instance
        entity = self._entity_class()

        # Resolvendo o join de conjuntos (se houver)
        with_conjunto = ""
        fields_conjunto = ""
        join_conjuntos = ""
        conjunto_map = {}
        if conjunto_type is not None:
            (
                join_conjuntos,
                with_conjunto,
                fields_conjunto,
                conjunto_map,
            ) = self._make_conjunto_sql(conjunto_type, entity, filters, conjunto_field)

        # Organizando o where dos filtros
        filters_where, filter_values_map = self._make_filters_sql(filters)

        # Montando a clausula dos fields vindos dos joins
        sql_join_fields, sql_join = self._make_joins_sql(joins_aux)

        if joins_aux:
            for join in joins_aux:
                for field in join.fields:
                    setattr(self._entity_class, f"{join.alias}_{field}", None)

        partial_exists_sql = ""
        if partial_exists_clause is not None:
            (
                partial_table_name,
                partial_base_field,
                partial_relation_field,
            ) = partial_exists_clause
            partial_exists_sql = f"""
            and exists (
                select 1
                from {partial_table_name} as partial_exists
                where partial_exists.{partial_relation_field} = t0.{partial_base_field}
            )
            """

        # Building query
        sql = f"""
        {with_conjunto}
        select
            {fields_conjunto}
            {self._sql_fields(fields)}
            {sql_join_fields}
        from
            {entity.get_table_name()} as t0
            {join_conjuntos}
            {sql_join}
        where
            t0.{key_field} = :id
            {filters_where}
            {partial_exists_sql}
        limit 10
        """
        values = {"id": id}
        values.update(filter_values_map)
        values.update(conjunto_map)

        # Running query
        resp = self._db.execute_query_to_model(sql, self._entity_class, **values)

        # Checking if ID was found
        if len(resp) <= 0:
            raise NotFoundException(
                f"{self._entity_class.__name__} com id {id} não encontrado."
            )

        # Verificando se foi encontrado mais de um registro para o ID passado
        if not override_data and len(resp) > 1:
            raise ConflictException(
                f"Encontrado mais de um registro do tipo {self._entity_class.__name__}, para o id {id}."
            )

        if not override_data:
            return resp[0]
        else:
            return resp



import datetime
import decimal
import re
import uuid

import unidecode

from typing import Any, Dict, List, Tuple

from nsj_rest_lib.entity.entity_base import EntityBase

from .dao_base_conjuntos import DAOBaseConjuntos

class DAOBaseSearch(DAOBaseConjuntos):

    def _make_search_sql(
        self,
        search_query: str,
        search_fields: List[str],
        entity: EntityBase,
    ) -> Tuple[Dict[str, Any], str]:
        """
        Monta a parte da cláusula where referente ao parâmetro search, bem como o mapa de
        valores para realizar a pesquisa (passando para a execução da query).

        Retorna uma tupla, onde a primeira posição é o mapa de valores, e a segunda a cláusula sql.
        """

        search_map = {}
        search_where = ""

        date_pattern = "(\d\d)/(\d\d)/((\d\d\d\d)|(\d\d))"
        int_pattern = "(\d+)"
        float_pattern = "(\d+((,|\.)\d+)?)"

        if search_fields is not None and search_query is not None:
            search_buffer = "false \n"
            for search_field in search_fields:
                search_str = search_query

                entity_field = entity.fields_map.get(search_field)
                if entity_field is None:
                    continue

                if (
                    entity_field.expected_type is datetime.datetime
                    or entity_field.expected_type is datetime.date
                ):
                    # Tratando da busca de datas
                    received_floats = re.findall(date_pattern, search_str)
                    cont = -1
                    for received_float in received_floats:
                        cont += 1

                        dia = int(received_float[0])
                        mes = int(received_float[0])
                        ano = received_float[0]
                        if len(ano) < 4:
                            ano = f"20{ano}"
                        ano = int(ano)

                        data_obj = None
                        try:
                            data_obj = datetime.date(ano, mes, dia)
                        except Exception:
                            continue

                        search_buffer += (
                            f" or t0.{search_field} = :shf_{search_field}_{cont} \n"
                        )
                        search_map[f"shf_{search_field}_{cont}"] = data_obj

                elif entity_field.expected_type is int:
                    # Tratando da busca de inteiros
                    search_str = re.sub(date_pattern, "", search_str)

                    received_floats = re.findall(int_pattern, search_str)
                    cont = -1
                    for received_float in received_floats:
                        cont += 1
                        valor = int(received_float[0])
                        valor_min = int(valor * 0.9)
                        valor_max = int(valor * 1.1)

                        search_buffer += f" or (t0.{search_field} >= :shf_{search_field}_{cont}_min and t0.{search_field} <= :shf_{search_field}_{cont}_max) \n"
                        search_map[f"shf_{search_field}_{cont}_min"] = valor_min
                        search_map[f"shf_{search_field}_{cont}_max"] = valor_max

                elif (
                    entity_field.expected_type is int
                    or entity_field.expected_type is decimal.Decimal
                ):
                    # Tratando da busca de floats e decimais
                    search_str = re.sub(date_pattern, "", search_str)

                    received_floats = re.findall(float_pattern, search_str)
                    cont = -1
                    for received_float in received_floats:
                        cont += 1
                        valor = float(received_float[0])
                        valor_min = valor * 0.9
                        valor_max = valor * 1.1

                        search_buffer += f" or (t0.{search_field} >= :shf_{search_field}_{cont}_min and t0.{search_field} <= :shf_{search_field}_{cont}_max) \n"
                        search_map[f"shf_{search_field}_{cont}_min"] = valor_min
                        search_map[f"shf_{search_field}_{cont}_max"] = valor_max

                elif (
                    entity_field.expected_type is str
                    or entity_field.expected_type is uuid
                ):
                    # Tratando da busca de strings e UUIDs
                    cont = -1
                    for palavra in search_str.split(" "):
                        if palavra == "":
                            continue

                        cont += 1
                        search_buffer += f" or upper(unaccent(CAST(t0.{search_field} AS varchar))) like upper(unaccent(:shf_{search_field}_{cont})) \n"
                        search_map[f"shf_{search_field}_{cont}"] = (
                            f"%{unidecode.unidecode(palavra)}%"
                        )

            search_where = f"""
            and (
                {search_buffer}
            )
            """

        return search_map, search_where



import typing as ty

from nsj_gcf_utils.json_util import convert_to_dumps

from nsj_rest_lib.dao.dao_base_save_by_function import _FunctionSQLBuilder
from nsj_rest_lib.dao.dao_base_util import DAOBaseUtil
from nsj_rest_lib.entity.function_type_base import FunctionTypeBase
from nsj_rest_lib.settings import get_logger


class DAOBaseFunction(DAOBaseUtil):
    def _call_function_with_type(
        self,
        function_object: FunctionTypeBase,
        function_name: str,
    ) -> list[dict]:
        """
        Executa uma função de banco que recebe um TYPE composto (FunctionType).
        Retorna a lista de registros (dict) gerada pela função.
        """

        function_type_class = function_object.__class__
        if not function_name:
            raise ValueError(
                f"É necessário informar o nome da função para {function_type_class.__name__}."
            )

        (
            relation_declarations,
            assignments,
            values_map,
        ) = _FunctionSQLBuilder(function_object).build()

        declarations_sql = "\n".join(
            f"    DECLARE {declaration}" for declaration in relation_declarations
        )

        assignments_sql = "\n".join(f"        {line}" for line in assignments)

        sql = f"""
        DO $DOFUNC$
            DECLARE VAR_TIPO {function_type_class.type_name};
{declarations_sql if declarations_sql else ''}
            DECLARE VAR_RETORNO JSONB;
        BEGIN
{assignments_sql}

            VAR_RETORNO = (
                SELECT coalesce(jsonb_agg(row_to_json(r)), '[]'::jsonb)
                FROM {function_name}(VAR_TIPO) r
            );
            PERFORM set_config('retorno.bloco', VAR_RETORNO::varchar, true);
        END $DOFUNC$;

        SELECT current_setting('retorno.bloco', true)::jsonb as retorno;
        """

        get_logger().debug(f"[RestLib Debug] Function SQL: {sql}")
        get_logger().debug(f"[RestLib Debug] Function Parameters: {values_map}")
        _, returning = self._db.execute_batch(sql, **values_map)

        if not returning:
            return []

        retorno = returning[0].get("retorno")
        if retorno is None:
            return []
        if isinstance(retorno, list):
            return retorno
        # Caso venha um único objeto jsonb
        return [retorno]

    def _call_function_raw(
        self,
        function_name: str,
        positional_values: list,
        named_values: dict[str, ty.Any] | None = None,
    ) -> list[dict]:
        """
        Executa uma função de banco passando parâmetros simples (sem TYPE composto).
        """
        named_values = named_values or {}

        values_map: dict[str, ty.Any] = {}
        placeholders = []

        for idx, value in enumerate(positional_values):
            placeholder = f"p_{idx}"
            placeholders.append(f":{placeholder}")
            values_map[placeholder] = convert_to_dumps(value)

        for key, value in named_values.items():
            safe_key = key.replace("-", "_")
            placeholders.append(f":{safe_key}")
            values_map[safe_key] = convert_to_dumps(value)

        args_sql = ", ".join(placeholders)
        sql = f"select * from {function_name}({args_sql});"

        returning = self._db.execute_query(sql, **values_map)

        return returning or []



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from nsj_rest_lib.descriptor.dto_sql_join_field import DTOSQLJoinField

from nsj_gcf_utils.sql_utils import SQLUtils

def montar_chave_map_sql_join(field: "DTOSQLJoinField") -> str:
    return f"{field.dto_type}____{field.entity_type}____{field.entity_relation_owner}____{field.join_type}"



import functools

from dataclasses import dataclass
from typing import Any, List, Optional, Set, Type

from nsj_rest_lib.descriptor.entity_field import EntityField
from nsj_rest_lib.entity.entity_base import EntityBase


@dataclass
class PartialEntityConfig:
    parent_entity: Type[EntityBase]
    extension_table_name: str
    parent_fields: Set[str]
    extension_fields: Set[str]


class Entity:
    def __init__(
        self,
        table_name: Optional[str] = None,
        pk_field: Optional[str] = None,
        default_order_fields: Optional[List[str]] = None,
        partial_of: Optional[Type[EntityBase]] = None,
        partial_table_name: Optional[str] = None,
    ) -> None:
        super().__init__()

        self.partial_of = partial_of

        parent_table_name = None
        parent_pk_field = None
        parent_default_order_fields = None
        if partial_of is not None:
            if not isinstance(partial_of, type) or not issubclass(
                partial_of, EntityBase
            ):
                raise ValueError(
                    "Configuração partial_of inválida: a classe informada deve ser subclasse de EntityBase."
                )

            parent_table_name = getattr(partial_of, "table_name", None)
            parent_pk_field = getattr(partial_of, "pk_field", None)
            parent_default_order_fields = list(
                getattr(partial_of, "default_order_fields", []) or []
            )

        extension_table_name = partial_table_name
        if partial_of is not None:
            if extension_table_name is None:
                extension_table_name = table_name

            if extension_table_name is None:
                raise ValueError(
                    "Entidades parciais devem informar o nome da tabela de extensão por meio de 'partial_table_name' ou 'table_name'."
                )

            if table_name is None:
                table_name = parent_table_name
            else:
                table_name = parent_table_name

            if pk_field is None:
                pk_field = parent_pk_field

            if default_order_fields is None:
                default_order_fields = parent_default_order_fields
        else:
            if table_name is None or pk_field is None or default_order_fields is None:
                raise ValueError(
                    "Parâmetros 'table_name', 'pk_field' e 'default_order_fields' são obrigatórios para entidades sem partial_of."
                )

        if default_order_fields is None:
            default_order_fields = []
        else:
            default_order_fields = list(default_order_fields)

        if pk_field is not None and pk_field not in default_order_fields:
            default_order_fields.append(pk_field)

        self.table_name = table_name
        self.pk_field = pk_field
        self.default_order_fields = default_order_fields
        self.partial_table_name = extension_table_name
        self.parent_default_order_fields = parent_default_order_fields

    def __call__(self, cls: object):
        """
        Tratando dos tipos de dados dos atributos, e criando os getters necessários.
        """

        # Mantém metadados da classe original
        functools.update_wrapper(self, cls)

        # Extendendo as propriedades por meio das extensões parciais
        parent_fields: Set[str] = set()
        if self.partial_of is not None:
            parent_fields = set(getattr(self.partial_of, "fields_map", {}).keys())
            parent_annotations = dict(
                getattr(self.partial_of, "__annotations__", {}) or {}
            )
            current_annotations = dict(getattr(cls, "__annotations__", {}) or {})
            merged_annotations = parent_annotations
            merged_annotations.update(current_annotations)
            cls.__annotations__ = merged_annotations

            for key in parent_annotations:
                if key not in cls.__dict__:
                    setattr(cls, key, getattr(self.partial_of, key, None))

            setattr(cls, "partial_entity_parent", self.partial_of)
            setattr(cls, "partial_extension_table", self.partial_table_name)
        else:
            setattr(cls, "partial_entity_parent", None)
            setattr(cls, "partial_extension_table", None)

        # Guardando o nome da tabela na classe
        self._check_class_attribute(cls, "table_name", self.table_name)

        # Guardando o nome do campo PK na classe
        self._check_class_attribute(cls, "pk_field", self.pk_field)

        # Guardando a lista default de ordenação, na classe
        self._check_class_attribute(
            cls, "default_order_fields", self.default_order_fields
        )

        # Creating fields_map in cls, if needed
        self._check_class_attribute(cls, "fields_map", {})

        # Iterando pelos atributos de classe
        for key, attr in cls.__dict__.items():
            atributo = None

            # Copiando o tipo a partir da anotação de tipo (se existir)
            if isinstance(attr, EntityField):
                atributo = attr
            elif key in cls.__annotations__:
                atributo = attr
                if not isinstance(attr, EntityField):
                    atributo = EntityField()

            if atributo:
                # Setting a better name to storage_name
                atributo.storage_name = f"{key}"
                atributo.name = f"{key}"

                # Guardando o type esperado
                if key in cls.__annotations__:
                    atributo.expected_type = cls.__annotations__[key]

                # Guardando o atributo no fields_map
                getattr(cls, "fields_map")[key] = atributo

        # Gravando as configurações d eextensão parcial
        if self.partial_of is not None:
            extension_fields = {
                key
                for key in getattr(cls, "fields_map").keys()
                if key not in parent_fields
            }
            setattr(
                cls,
                "partial_entity_config",
                PartialEntityConfig(
                    parent_entity=self.partial_of,
                    extension_table_name=self.partial_table_name,
                    parent_fields=parent_fields,
                    extension_fields=extension_fields,
                ),
            )
        else:
            setattr(cls, "partial_entity_config", None)

        return cls

    def _check_class_attribute(self, cls: object, attr_name: str, default_value: Any):
        """
        Add attribute "attr_name" in class "cls", if not exists.
        """

        if attr_name not in cls.__dict__:
            setattr(cls, attr_name, default_value)



import enum
import typing as ty

from nsj_rest_lib.entity.entity_base import EntityBase
from nsj_rest_lib.entity.function_type_base import (
    InsertFunctionTypeBase,
    UpdateFunctionTypeBase,
)

from .dto_field import DTOField

if ty.TYPE_CHECKING is True:
    from nsj_rest_lib.dto.dto_base import DTOBase
    from .dto_left_join_field import EntityRelationOwner

    pass

T = ty.TypeVar('T')


class OTORelationType(enum.IntEnum):
    """The enum for Relation Type for One to One relations."""

    COMPOSITION = 0
    AGGREGATION = 1
    pass


# pylint: disable=too-many-instance-attributes
class DTOOneToOneField:
    _ref_counter = 0

    expected_type: ty.Type['DTOBase']
    relation_type: OTORelationType
    field: DTOField
    entity_relation_owner: 'EntityRelationOwner'
    not_null: bool
    resume: bool
    partition_data: bool
    entity_field: str
    relation_field: str
    insert_function_field: str
    update_function_field: str
    validator: ty.Optional[ty.Callable[..., ty.Any]]
    description: str
    convert_to_function: ty.Optional[ty.Callable[..., ty.Any]]

    def __init__(
        self,
        entity_type: ty.Type[EntityBase],
        relation_type: OTORelationType,
        resume: bool = False,
        entity_field: ty.Optional[str] = None,
        entity_relation_owner: 'EntityRelationOwner' = 'self',  # type: ignore
        not_null: bool = False,
        partition_data: bool = False,
        validator: ty.Optional[ty.Callable[['DTOOneToOneField', T], T]] = None,
        description: str = '',
        insert_function_field: ty.Optional[str] = None,
        insert_function_type: ty.Optional[ty.Type[InsertFunctionTypeBase]] = None,
        update_function_field: ty.Optional[str] = None,
        update_function_type: ty.Optional[ty.Type[UpdateFunctionTypeBase]] = None,
        convert_to_function: ty.Optional[ty.Callable[..., ty.Any]] = None,
        get_function_field: ty.Optional[str] = None,
        delete_function_field: ty.Optional[str] = None,
        relation_field: ty.Optional[str] = None,
    ):
        """Descriptor used for One to One relations.
        ---------
        Glossary:
        ---------
        - Current DTO: Refers to the DTO that this field is a part of.
        - Related DTO: Refers to the DTO in the annotation of this field.


        -----
        NOTE:
        -----
        At the moment only `entity_relation_owner=EntityRelationOwner.SELF` is
            supported.

        -----------
        Parameters:
        -----------

        - entity_type: Entity type of the `Related DTO`
            (must be a subclasse from EntityBase).

        - relation_type: The type of relation of this field, one of:
            - OTORelationType.COMPOSITION:
                - During POST requests, it attempts to insert the data into the
                    `entity_type` table.
                - During PUT or PATCH requests, it attempts to update the
                    existing records in the `entity_type` table.
            - OTORelationType.AGGREGATION:
                - This type does not interact with the `entity_type` table.
                - It is relevant only in POST, PUT, or PATCH requests if the
                    `relation_field` is `None`. In this case, the value in the
                    `pk_field` of the `Related DTO` will be used in place of
                    the object.

        - relation_field: Field name in the `entity_type` to use in the `on` of
            the `join` query. Defaults to the `pk_field` of the `entity_type`.

        - resume: Indicates if on GET requests the non expanded value should be
            always returned.

        - entity_field: The name of the field in the Entity of the `Current DTO`.
            If `None` will use the name of the field in the `Current DTO`.

        - insert_function_field: Nome opcional do campo correspondente no InsertFunctionType (default: nome do campo no DTO).

        - update_function_field: Nome opcional do campo correspondente no UpdateFunctionType (default: herdado do campo de insert).

        - entity_relation_owner: Indicates which entity contain the
            `relation_field`, it must be one of:
                - EntityRelationField.SELF: The `relation_field` is part of the
                    `Current DTO`.
                - EntityRelationField.OTHER: The `relation_field` is part os the
                    `Related DTO`.

        - not_null: If the field can not be `None`. Only relevant in POST, PUT
            or PATCH requests.

        - partition_data: If the propertie is obligatory when listing(GET/LIST)
            including on relations.

        - validator: Function that receives the instance of this class and the
            value to be checked and returns it. For validation erros MUST throw
            ValueError. Errors are only honored on POST, PUT or PATCH requests.
            When `OTORelationType.AGGREGATION` the value passed will be the value
            in the `pk_field` of the `Related DTO`.

        - description: Description of this field that can be used in
            documentation.

        - convert_to_function: Função usada para converter o valor antes de popular o InsertFunctionType. Recebe (valor, dict_com_valores_do_dto) e deve retornar um dicionário com os campos/resultados a atribuir.

        - get_function_field: Nome do campo equivalente no Get/ListFunctionType (default: o próprio nome do campo no DTO).

        - delete_function_field: Nome do campo equivalente no DeleteFunctionType (default: o próprio nome do campo no DTO).
        """
        self.entity_type = entity_type
        self.relation_type = relation_type
        self.resume = resume
        self.entity_field = entity_field or ''
        self.insert_function_field = insert_function_field
        self.insert_function_type = insert_function_type
        self.update_function_field = update_function_field
        self.update_function_type = update_function_type
        self.entity_relation_owner = entity_relation_owner
        self.not_null = not_null
        self.partition_data = partition_data
        self.validator = validator
        self.description = description
        self.convert_to_function = convert_to_function
        self.get_function_field = get_function_field
        self.delete_function_field = delete_function_field
        self.relation_field = relation_field or ''

        self.name = None
        self.expected_type = ty.cast(ty.Type['DTOBase'], type)

        self.storage_name = (
            f"_{self.__class__.__name__}#{self.__class__._ref_counter}"
        )
        self.__class__._ref_counter += 1

        if (
            self.insert_function_type is not None
            and not issubclass(self.insert_function_type, InsertFunctionTypeBase)
        ):
            raise ValueError(
                "insert_function_type deve herdar de InsertFunctionTypeBase."
            )

        if (
            self.update_function_type is not None
            and not issubclass(self.update_function_type, UpdateFunctionTypeBase)
        ):
            raise ValueError(
                "update_function_type deve herdar de UpdateFunctionTypeBase."
            )

        # NOTE: To support EntityRelationOwner.OTHER you will have to modify
        #           `_retrieve_one_to_one_fields in ServiceBase`. do NOT forget
        #           to change the documentation.
        assert (
            self.entity_relation_owner == 'self'
        ), "At the moment only `EntityRelationOwner.SELF` is supported."

        assert issubclass(self.entity_type, EntityBase), (
            f"Argument `entity_type` of `DTOOneToOneField` HAS to be"
            f" a `EntityBase`. Is {repr(self.entity_type)}."
        )

        if self.relation_field == '':
            self.relation_field = self.entity_type.pk_field
            pass

        assert self.relation_field in self.entity_type.fields_map, (
            f"Argument `relation_field` of `DTOOneToOneField` HAS to be"
            f" a field of `{repr(self.entity_type)}`. Is {repr(self.relation_field)}."
        )

        pass

    def __get__(self, instance: ty.Optional['DTOBase'], owner: ty.Any):
        if instance is None:
            return self
        return instance.__dict__[self.storage_name]

    def __set__(
        self, instance: ty.Optional['DTOBase'], value: ty.Optional[ty.Any]
    ) -> None:
        escape_validator: bool = False
        if (
            'escape_validator' in instance.__dict__
            and instance.__dict__['escape_validator'] is True
        ):
            escape_validator = True
            pass

        try:
            if self.not_null is True and value is None:
                raise ValueError(f"{self.storage_name} deve ser preenchido.")

            if value is None:
                instance.__dict__[self.storage_name] = None
                return

            if self.relation_type == OTORelationType.AGGREGATION:
                if escape_validator is True:
                    if isinstance(value, dict):
                        value = self.expected_type(**value)
                        pass
                else:
                    if isinstance(value, self.expected_type):
                        value = getattr(value, self.relation_field, None)
                    elif isinstance(value, dict):
                        value = value.get(self.relation_field, None)
                        pass
                    pass

                # NOTE: At this moment if the relation_field in the entity has
                #           a diferent name from the field on the DTO we disable
                #           validation, because at the moment there is no easy
                #           way to get a DTOField from the entity field name
                if isinstance(value, self.expected_type) and hasattr(value, self.relation_field):
                    relation_value = getattr(value, self.relation_field)
                    if self.field.use_default_validator:
                        # NOTE: This may throw
                        relation_value = self.field.validate(
                            self.field, relation_value, instance
                        )
                        pass

                    if self.field.validator is not None:
                        # NOTE: This may throw
                        relation_value = self.field.validator(
                            self.field, relation_value
                        )
                        pass

                    setattr(value, self.relation_field, relation_value)
                else:
                    if self.field.use_default_validator:
                        # NOTE: This may throw
                        value = self.field.validate(self.field, value, instance)
                        pass

                    if self.field.validator is not None:
                        # NOTE: This may throw
                        value = self.field.validator(self.field, value)
                        pass
                    pass
            else:  # self.relation_type == RelationType.COMPOSITION
                if isinstance(value, dict):
                    value = self.expected_type(**value)  # NOTE: This may throw
                    pass

                if not isinstance(value, self.expected_type):
                    raise ValueError(
                        f"{self.storage_name} deve ser do tipo: {self.expected_type}"
                    )
                if self.validator is not None:
                    value = self.validator(self, value)  # NOTE: This may throw
                    pass
                pass
        except ValueError:
            if escape_validator is False:
                raise
            pass

        instance.__dict__[self.storage_name] = value
        pass

    def get_insert_function_field_name(self) -> str:
        if self.insert_function_field is not None:
            return self.insert_function_field
        return self.name

    def get_update_function_field_name(self) -> str:
        if self.update_function_field is not None:
            return self.update_function_field
        return self.get_insert_function_field_name()

    def get_function_field_name(self, operation: str) -> str:
        if operation in ("get", "list"):
            if self.get_function_field is not None:
                return self.get_function_field
            return self.name
        if operation == "delete":
            if self.delete_function_field is not None:
                return self.delete_function_field
            return self.name
        if operation == "update":
            return self.get_update_function_field_name()
        return self.get_insert_function_field_name()

    def get_function_type(self, operation: str):
        if operation == "update" and self.update_function_type is not None:
            return self.update_function_type
        return self.insert_function_type



# from .dto_field import DTOField, DTOFieldFilter, DTOAutoIncrementField
# from .dto_aggregator import DTOAggregator
# from .dto_one_to_one_field import DTOOneToOneField, OTORelationType

# __all__ = [
#     "DTOField",
#     "DTOFieldFilter",
#     "DTOAutoIncrementField",
#     "DTOAggregator",
#     "DTOOneToOneField",
#     "OTORelationType",
# ]



import os
import typing as ty

from flask import request
from typing import Callable, Type

from nsj_audit_lib.util.audit_config import AuditConfig
from nsj_gcf_utils.json_util import json_dumps, JsonLoadException
from nsj_gcf_utils.rest_error_util import format_json_error

from nsj_rest_lib.controller.controller_util import DEFAULT_RESP_HEADERS
from nsj_rest_lib.controller.route_base import RouteBase
from nsj_rest_lib.dao.dao_base import DAOBase
from nsj_rest_lib.dto.dto_base import DTOBase
from nsj_rest_lib.dto.queued_data_dto import QueuedDataDTO
from nsj_rest_lib.entity.entity_base import EntityBase
from nsj_rest_lib.entity.function_type_base import InsertFunctionTypeBase
from nsj_rest_lib.exception import (
    MissingParameterException,
    ConflictException,
)
from nsj_rest_lib.injector_factory_base import NsjInjectorFactoryBase
from nsj_rest_lib.settings import get_logger


class PostRoute(RouteBase):
    def __init__(
        self,
        url: str,
        http_method: str,
        dto_class: DTOBase,
        entity_class: EntityBase,
        dto_response_class: DTOBase = None,
        injector_factory: NsjInjectorFactoryBase = NsjInjectorFactoryBase,
        service_name: str = None,
        handle_exception: Callable = None,
        custom_before_insert: Callable = None,
        custom_after_insert: Callable = None,
        retrieve_after_insert: bool = False,
        custom_json_response: bool = False,
        insert_function_type_class: Type[InsertFunctionTypeBase] | None = None,
        insert_function_name: str | None = None,
        audit_config: AuditConfig | None = None,
    ):
        super().__init__(
            url=url,
            http_method=http_method,
            dto_class=dto_class,
            entity_class=entity_class,
            dto_response_class=dto_response_class,
            injector_factory=injector_factory,
            service_name=service_name,
            handle_exception=handle_exception,
            audit_config=audit_config,
        )
        self.custom_before_insert = custom_before_insert
        self.custom_after_insert = custom_after_insert
        self.retrieve_after_insert = retrieve_after_insert
        self.custom_json_response = custom_json_response
        self._insert_function_type_class = insert_function_type_class
        self._insert_function_name = insert_function_name

        if self._insert_function_type_class is not None and not issubclass(
            self._insert_function_type_class, InsertFunctionTypeBase
        ):
            raise ValueError(
                "A classe informada em insert_function_type_class deve herdar de InsertFunctionTypeBase."
            )

    def _get_service(self, factory: NsjInjectorFactoryBase):
        """
        Sobrescreve o _get_service padrão para permitir configurar
        o InsertFunctionType e o nome da função diretamente no Service.
        """

        if self._service_name is not None:
            return factory.get_service_by_name(self._service_name)

        from nsj_rest_lib.service.service_base import ServiceBase

        return ServiceBase(
            factory,
            DAOBase(factory.db_adapter(), self._entity_class),
            self._dto_class,
            self._entity_class,
            self._dto_response_class,
            insert_function_type_class=self._insert_function_type_class,
            insert_function_name=self._insert_function_name,
        )

    def _partition_filters(self, data):
        # Montando os filtros de particao de dados
        partition_filters = {}

        for field in data.partition_fields:
            value = getattr(data, field)
            if value is None:
                raise MissingParameterException(field)
            elif value is not None:
                partition_filters[field] = value

        return partition_filters

    def handle_request(
        self,
        query_args: dict[str, any] = None,
        body: dict[str, any] = None,
        **kwargs: ty.Any,
    ):
        """
        Tratando requisições HTTP Post para inserir uma instância de uma entidade.
        """

        try:
            # Recuperando os dados do corpo da requisição
            if os.getenv("ENV", "").lower() != "erp_sql":
                request_data = request.json
            else:
                request_data = body

            if not isinstance(request_data, list):
                request_data = [request_data]

            args = (
                request.args
                if os.getenv("ENV", "").lower() != "erp_sql"
                else query_args or {}
            )
            retrieve_fields = (
                RouteBase.parse_fields(self._dto_class, args.get("fields"))
                if self.retrieve_after_insert
                else None
            )

            data_pack = []
            lst_data = []
            partition_filters = None
            for item in request_data:
                if len(kwargs) >= 0:
                    item.update(kwargs)

                # Convertendo os dados para o DTO
                data = self._dto_class(validate_read_only=True, **item)

                # Montando os filtros de particao de dados
                if partition_filters is None:
                    partition_filters = self._partition_filters(data)

                data_pack.append(data)

            if partition_filters is None:
                if len(kwargs) > 0:
                    partition_filters = kwargs.copy()
            else:
                partition_filters.update(kwargs)

            # Construindo os objetos
            service = self._get_service(self.get_injector_factory())

            if len(data_pack) == 1:
                # Chamando o service (método insert)
                data = service.insert(
                    dto=data,
                    aditional_filters=partition_filters,
                    custom_before_insert=self.custom_before_insert,
                    custom_after_insert=self.custom_after_insert,
                    retrieve_after_insert=self.retrieve_after_insert,
                    function_name=self._insert_function_name,
                    custom_json_response=self.custom_json_response,
                    retrieve_fields=retrieve_fields,
                )

                if data is not None:
                    # Verificando se houve um enfileiramento (pelo custom_after_insert)
                    if isinstance(data, QueuedDataDTO):
                        queued_data: QueuedDataDTO = data
                        resp_headers = {
                            **DEFAULT_RESP_HEADERS,
                            "Location": queued_data.status_url,
                        }
                        return ("", 202, resp_headers)

                    if self.custom_json_response and (
                        isinstance(data, dict)
                        or (
                            isinstance(data, list)
                            and (not data or not hasattr(data[0], "convert_to_dict"))
                        )
                    ):
                        return (json_dumps(data), 200, {**DEFAULT_RESP_HEADERS})

                    # Convertendo para o formato de dicionário (permitindo omitir campos do DTO)
                    lst_data.append(data.convert_to_dict(retrieve_fields))
            else:
                data = service.insert_list(
                    dtos=data_pack,
                    aditional_filters=partition_filters,
                    custom_before_insert=self.custom_before_insert,
                    custom_after_insert=self.custom_after_insert,
                    retrieve_after_insert=self.retrieve_after_insert,
                    function_name=self._insert_function_name,
                    custom_json_response=self.custom_json_response,
                    retrieve_fields=retrieve_fields,
                )

                if (
                    self.custom_json_response
                    and isinstance(data, list)
                    and (not data or not hasattr(data[0], "convert_to_dict"))
                ):
                    return (json_dumps(data), 200, {**DEFAULT_RESP_HEADERS})

                if data is not None or not len(data) > 0:
                    # Convertendo para o formato de dicionário (permitindo omitir campos do DTO)
                    lst_data = [item.convert_to_dict(retrieve_fields) for item in data]

            if len(lst_data) == 1:
                # Retornando a resposta da requisição
                return (json_dumps(lst_data[0]), 200, {**DEFAULT_RESP_HEADERS})

            if len(lst_data) > 1:
                # Retornando a resposta da requisição
                return (json_dumps(lst_data), 200, {**DEFAULT_RESP_HEADERS})

            # Retornando a resposta da requisição
            return ("", 201, {**DEFAULT_RESP_HEADERS})
        except JsonLoadException as e:
            get_logger().warning(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (format_json_error(e), 400, {**DEFAULT_RESP_HEADERS})
        except MissingParameterException as e:
            get_logger().warning(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (format_json_error(e), 400, {**DEFAULT_RESP_HEADERS})
        except ValueError as e:
            get_logger().warning(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (format_json_error(e), 400, {**DEFAULT_RESP_HEADERS})
        except ConflictException as e:
            get_logger().warning(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (format_json_error(e), 409, {**DEFAULT_RESP_HEADERS})
        except Exception as e:
            get_logger().exception(e)
            if self._handle_exception is not None:
                return self._handle_exception(e)
            else:
                return (
                    format_json_error(f"Erro desconhecido: {e}"),
                    500,
                    {**DEFAULT_RESP_HEADERS},
                )
